# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import contextlib
import datetime
import functools
import importlib
import re
import signal
import urllib.parse as urlparse

from cdpcli import LIST_TYPE
from cdpcli import MAP_TYPE
from cdpcli import OBJECT_TYPE
from cdpcli.compat import compat_tzlocal
from cdpcli.compat import OrderedDict
import dateutil.parser
from dateutil.tz import tzutc

# These are chars that do not need to be urlencoded based on rfc2986, section 2.3.
SAFE_CHARS = '-._~'


def import_module(module_name):
    try:
        return importlib.import_module(module_name)
    except ImportError:
        return None


def get_extension_registers(name):
    module = import_module('cdpcli.extensions.%s.register' % name)
    if module is None:
        module = import_module('cdpcli.extensions.%s' % name)
    if module is None:
        return None, None

    register = getattr(module, 'register', None)
    if register is not None:
        return register, None

    register_ext = getattr(module, 'register_extension', None)
    register_cmd = getattr(module, 'register_command', None)
    return register_ext, register_cmd


def is_absolute_url(url):
    return bool(urlparse.urlparse(url).netloc)


def get_service_module_name(service_model):
    name = service_model.service_name
    name = name.replace('Cloudera', '')
    name = name.replace('CDP', '')
    name = re.sub('\\W+', '', name)
    return name


def json_encoder(obj):
    """JSON encoder that formats datetimes as ISO8601 format."""
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()
    else:
        return obj


class CachedProperty(object):
    def __init__(self, fget):
        self._fget = fget

    def __get__(self, obj, cls):
        if obj is None:
            return self
        else:
            computed_value = self._fget(obj)
            obj.__dict__[self._fget.__name__] = computed_value
            return computed_value


def instance_cache(func):
    """Method decorator for caching method calls to a single instance.

    **This is not a general purpose caching decorator.**

    In order to use this, you *must* provide an ``_instance_cache``
    attribute on the instance.

    This decorator is used to cache method calls.  The cache is only
    scoped to a single instance though such that multiple instances
    will maintain their own cache.  In order to keep things simple,
    this decorator requires that you provide an ``_instance_cache``
    attribute on your instance.

    """
    func_name = func.__name__

    @functools.wraps(func)
    def _cache_guard(self, *args, **kwargs):
        cache_key = (func_name, args)
        if kwargs:
            kwarg_items = tuple(sorted(kwargs.items()))
            cache_key = (func_name, args, kwarg_items)
        result = self._instance_cache.get(cache_key)
        if result is not None:
            return result
        result = func(self, *args, **kwargs)
        self._instance_cache[cache_key] = result
        return result

    return _cache_guard


def parse_timestamp(value):
    """Parse a timestamp into a datetime object.

    Supported formats:

        * iso8601
        * rfc822
        * epoch (value is an integer)

    This will return a ``datetime.datetime`` object.

    """
    if isinstance(value, (int, float)):
        return datetime.datetime.fromtimestamp(value, compat_tzlocal())
    else:
        try:
            return datetime.datetime.fromtimestamp(float(value), compat_tzlocal())
        except (TypeError, ValueError):
            pass
    try:
        return dateutil.parser.parse(value,
                                     tzinfos={'GMT': tzutc()})
    except (TypeError, ValueError) as e:
        raise ValueError('Invalid timestamp "%s": %s' % (value, e))


def parse_to_aware_datetime(value):
    """Converted the passed in value to a datetime object with tzinfo.

    This function can be used to normalize all timestamp inputs.  This
    function accepts a number of different types of inputs, but
    will always return a datetime.datetime object with time zone
    information.

    The input param ``value`` can be one of several types:

        * A datetime object (both naive and aware)
        * An integer representing the epoch time (can also be a string
          of the integer, i.e '0', instead of 0).  The epoch time is
          considered to be UTC.
        * An iso8601 formatted timestamp.  This does not need to be
          a complete timestamp, it can contain just the date portion
          without the time component.

    The returned value will be a datetime object that will have tzinfo.
    If no timezone info was provided in the input value, then UTC is
    assumed, not local time.

    """
    # This is a general purpose method that handles several cases of
    # converting the provided value to a string timestamp suitable to be
    # serialized to an http request. It can handle:
    # 1) A datetime.datetime object.
    if isinstance(value, datetime.datetime):
        datetime_obj = value
    else:
        # 2) A string object that's formatted as a timestamp.
        #    We document this as being an iso8601 timestamp, although
        #    parse_timestamp is a bit more flexible.
        datetime_obj = parse_timestamp(value)
    if datetime_obj.tzinfo is None:
        # I think a case would be made that if no time zone is provided,
        # we should use the local time.  However, to restore backwards
        # compat, the previous behavior was to assume UTC, which is
        # what we're going to do here.
        datetime_obj = datetime_obj.replace(tzinfo=tzutc())
    else:
        datetime_obj = datetime_obj.astimezone(tzutc())
    return datetime_obj


@contextlib.contextmanager
def ignore_ctrl_c():
    original = signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        yield
    finally:
        signal.signal(signal.SIGINT, original)


def datetime2timestamp(dt, default_timezone=None):
    """Calculate the timestamp based on the given datetime instance.

    :type dt: datetime
    :param dt: A datetime object to be converted into timestamp
    :type default_timezone: tzinfo
    :param default_timezone: If it is provided as None, we treat it as tzutc().
                             But it is only used when dt is a naive datetime.
    :returns: The timestamp
    """
    epoch = datetime.datetime(1970, 1, 1)
    if dt.tzinfo is None:
        if default_timezone is None:
            default_timezone = tzutc()
        dt = dt.replace(tzinfo=default_timezone)
    d = dt.replace(tzinfo=None) - dt.utcoffset() - epoch
    if hasattr(d, "total_seconds"):
        return d.total_seconds()  # Works in Python 2.7+
    return (d.microseconds + (d.seconds + d.days * 24 * 3600) * 10 ** 6) / 10 ** 6


class ArgumentGenerator(object):
    """Generate sample input based on a shape model.

    This class contains a ``generate_skeleton`` method that will take
    an input shape (created from ``cdpcli.model``) and generate
    a sample dictionary corresponding to the input shape.

    The specific values used are place holder values. For strings an
    empty string is used, for numbers 0 or 0.0 is used. For datetime a date in
    ISO8601 is used. The intended usage of this class is to generate the *shape*
    of the input object. In the future we might take the defaults from the
    model.

    This can be useful for operations that have complex input shapes.
    This allows a user to just fill in the necessary data instead of
    worrying about the specific object of the input arguments.

    Example usage::

        clidriver = CLIDriver
        ddb = clidriver.get_service_model('dataeng')
        arg_gen = ArgumentGenerator()
        sample_input = arg_gen.generate_skeleton(
            ddb.operation_model('createAWSCluster').input_shape)
        print("Sample input for dataeng.createAWSCluster: %s" % sample_input)

    """

    def __init__(self):
        pass

    def generate_skeleton(self, shape):
        if shape.type_name == OBJECT_TYPE:
            return self._generate_type_object(shape)
        elif shape.type_name == LIST_TYPE:
            return self._generate_type_array(shape)
        elif shape.type_name == MAP_TYPE:
            return self._generate_type_map(shape)
        elif shape.type_name == 'string':
            return ''
        elif shape.type_name in ['integer']:
            return 0
        elif shape.type_name in ['float', 'double']:
            return 0.0
        elif shape.type_name == 'boolean':
            return True
        elif shape.type_name == 'datetime':
            return '2002-10-02T13:00:00+00:00'
        elif shape.type_name == 'blob':
            return ''
        else:
            raise Exception("Unknown shape type: %s" % shape.type_name)

    def _generate_type_object(self, shape):
        skeleton = OrderedDict()
        for member_name, member_shape in shape.members.items():
            skeleton[member_name] = self.generate_skeleton(member_shape)
        return skeleton

    def _generate_type_array(self, shape):
        # For list elements we've arbitrarily decided to return the first
        # element for the skeleton list.
        return [
            self.generate_skeleton(shape.member),
        ]

    def _generate_type_map(self, shape):
        key_shape = shape.key
        value_shape = shape.value
        assert key_shape.type_name == 'string'
        return OrderedDict([
            ('KeyName', self.generate_skeleton(value_shape)),
        ])



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2020 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import copy

from cdpcli.compat import OrderedDict
from cdpcli.endpoint import DEFAULT_TIMEOUT


class Config(object):
    """Advanced configuration for CDP CLI clients.

    :type connect_timeout: float or int
    :param connect_timeout: The time in seconds till a timeout exception is
        thrown when attempting to make a connection. The default is 60
        seconds.
    :type read_timeout: float or int
    :param read_timeout: The time in seconds till a timeout exception is
        thrown when attempting to read from a connection. The default is
        60 seconds.
    """

    OPTION_DEFAULTS = OrderedDict([
        ('connect_timeout', DEFAULT_TIMEOUT),
        ('read_timeout', DEFAULT_TIMEOUT),
    ])

    def __init__(self, *args, **kwargs):
        self._user_provided_options = self._record_user_provided_options(
            args, kwargs)

        # Merge the user_provided options onto the default options
        config_vars = copy.copy(self.OPTION_DEFAULTS)
        config_vars.update(self._user_provided_options)

        # Set the attributes based on the config_vars
        for key, value in config_vars.items():
            setattr(self, key, value)

    def _record_user_provided_options(self, args, kwargs):
        option_order = list(self.OPTION_DEFAULTS)
        user_provided_options = {}

        # Iterate through the kwargs passed through to the constructor and
        # map valid keys to the dictionary
        for key, value in kwargs.items():
            if key in self.OPTION_DEFAULTS:
                user_provided_options[key] = value
            # The key must exist in the available options
            else:
                raise TypeError(
                    'Got unexpected keyword argument \'%s\'' % key)

        # The number of args should not be longer than the allowed
        # options
        if len(args) > len(option_order):
            raise TypeError(
                'Takes at most %s arguments (%s given)' % (
                    len(option_order), len(args)))

        # Iterate through the args passed through to the constructor and map
        # them to appropriate keys.
        for i, arg in enumerate(args):
            # If it a kwarg was specified for the arg, then error out
            if option_order[i] in user_provided_options:
                raise TypeError(
                    'Got multiple values for keyword argument \'%s\'' % (
                        option_order[i]))
            user_provided_options[option_order[i]] = arg

        return user_provided_options

    def merge(self, other_config):
        """Merges the config object with another config object
        This will merge in all non-default values from the provided config
        and return a new config object
        :type other_config: botocore.config.Config
        :param other config: Another config object to merge with. The values
            in the provided config object will take precedence in the merging
        :returns: A config object built from the merged values of both
            config objects.
        """
        # Make a copy of the current attributes in the config object.
        config_options = copy.copy(self._user_provided_options)

        # Merge in the user provided options from the other config
        config_options.update(other_config._user_provided_options)

        # Return a new config object with the merged properties.
        return Config(**config_options)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging

from cdpcli import model
from cdpcli import xform_name
from cdpcli.arguments import BaseCLIArgument
from cdpcli.exceptions import PaginationError


LOG = logging.getLogger(__name__)


MAX_ITEMS_ARG_NAME = 'max-items'
MAX_ITEMS_HELP_FORMAT = """
<p>The total number of items to return. If the total number of items available
is more than the value specified in max-items then a <code>nextToken</code> will
be provided in the output that you can use to resume pagination. This
<code>nextToken</code> response element should <b>not</b> be used directly
outside of the CDP CLI. This option cannnot be combined with the
<code>no-paginate</code> CLI option. If no max-items value is specified, then
a default value of %d is used.</p>
"""

STARTING_TOKEN_ARG_NAME = 'starting-token'
STARTING_TOKEN_HELP = """
<p> A token to specify where to start paginating. This is the
<code>nextToken</code> from a previously truncated response.</p>
"""

PAGE_SIZE_ARG_NAME = 'page-size'
PAGE_SIZE_HELP = """
<p>The size of each page. Generally this does not need to be set and the
default page size used is appropriate.<p>
"""

PAGING_ARG_NAMES = [STARTING_TOKEN_ARG_NAME,
                    MAX_ITEMS_ARG_NAME,
                    PAGE_SIZE_ARG_NAME]

NORMALIZED_PAGING_ARGS = [STARTING_TOKEN_ARG_NAME.replace('-', '_'),
                          MAX_ITEMS_ARG_NAME.replace('-', '_'),
                          PAGE_SIZE_ARG_NAME.replace('-', '_')]


def add_pagination_params(operation_model, argument_table):
    unify_paging_params(operation_model, argument_table)


def add_paging_description(help_command):
    # This customization is only applied to the description of
    # Operations, so we must filter out all other operation types.
    if not isinstance(help_command.obj, model.OperationModel):
        return
    operation_model = help_command.obj
    if not operation_model.can_paginate:
        return
    help_command.doc.style.new_paragraph()
    help_command.doc.writeln(
        ('``%s`` is a paginated operation. Multiple API calls may be issued '
         'in order to retrieve the entire data set of results. You can '
         'disable pagination by providing the ``--no-paginate`` argument.')
        % help_command.name)


def unify_paging_params(operation_model, argument_table):
    if not operation_model.can_paginate:
        return

    LOG.debug("Modifying paging parameters for operation: %s" %
              operation_model.name)
    _remove_existing_paging_arguments(argument_table, operation_model)

    arg = PageArgument(argument_table,
                       MAX_ITEMS_ARG_NAME,
                       MAX_ITEMS_HELP_FORMAT
                       % operation_model.paging_default_max_items,
                       parse_type='integer',
                       serialized_name='MaxItems')
    if MAX_ITEMS_ARG_NAME in argument_table:
        del argument_table[MAX_ITEMS_ARG_NAME]
    argument_table[MAX_ITEMS_ARG_NAME] = arg

    arg = PageArgument(argument_table,
                       STARTING_TOKEN_ARG_NAME,
                       STARTING_TOKEN_HELP,
                       parse_type='string',
                       serialized_name='StartingToken')
    if STARTING_TOKEN_ARG_NAME in argument_table:
        del argument_table[STARTING_TOKEN_ARG_NAME]
    argument_table[STARTING_TOKEN_ARG_NAME] = arg

    arg = PageArgument(argument_table,
                       PAGE_SIZE_ARG_NAME,
                       PAGE_SIZE_HELP,
                       parse_type='integer',
                       serialized_name='PageSize')
    if PAGE_SIZE_ARG_NAME in argument_table:
        del argument_table[PAGE_SIZE_ARG_NAME]
    argument_table[PAGE_SIZE_ARG_NAME] = arg


def check_should_enable_pagination(argument_table, operation_model, parsed_args,
                                   parsed_globals):
    for token in _get_all_cli_input_tokens(operation_model):
        py_name = token.replace('-', '_')
        if getattr(parsed_args, py_name) is not None and \
                py_name not in NORMALIZED_PAGING_ARGS:
            # The user has specified a manual (undocumented) pagination arg.
            # We need to automatically turn pagination off.
            LOG.debug("User has specified a manual pagination arg: %s. "
                      "Automatically setting --no-paginate." % token)
            parsed_globals.paginate = False

            # Because pagination is now disabled, there's a chance that
            # we were shadowing arguments.  For example, we inject a
            # --max-items argument in unify_paging_params().  If the
            # the operation also provides its own MaxItems (which we
            # expose as --max-items) then our custom pagination arg
            # was shadowing the customers arg.  When we turn pagination
            # off we need to put back the original argument which is
            # what we're doing here.
            for shadowed_arg in _get_all_shadowed_paging_args(argument_table):
                argument_table[shadowed_arg.name] = \
                    shadowed_arg.shadowed_argument

    if not parsed_globals.paginate:
        ensure_paging_params_not_set(argument_table, parsed_args)


def ensure_paging_params_not_set(argument_table, parsed_args):
    shadowed_params = []
    for shadowed_arg in _get_all_shadowed_paging_args(argument_table):
        shadowed_params.extend(shadowed_arg.name.replace('-', '_'))

    params_used = [p for p in NORMALIZED_PAGING_ARGS if
                   p not in shadowed_params and getattr(parsed_args, p, None)]

    if len(params_used) > 0:
        converted_params = ', '.join(
            ["--" + p.replace('_', '-') for p in params_used])
        raise PaginationError(
            message="Cannot specify --no-paginate along with pagination "
                    "arguments: %s" % converted_params)


def _get_all_shadowed_paging_args(argument_table):
    for paging_arg_name in PAGING_ARG_NAMES:
        paging_arg = argument_table.get(paging_arg_name, None)
        if paging_arg is not None and isinstance(paging_arg, PageArgument) and \
           paging_arg.shadowed_argument is not None:
            yield paging_arg


def _remove_existing_paging_arguments(argument_table, operation_model):
    for token in _get_all_cli_input_tokens(operation_model):
        argument_table[token]._UNDOCUMENTED = True


def _get_all_cli_input_tokens(operation_model):
    if operation_model.can_paginate:
        yield xform_name(operation_model.paging_input_token, '-')
        yield xform_name(operation_model.paging_page_size, '-')


class PageArgument(BaseCLIArgument):
    type_map = {
        'string': str,
        'integer': int,
    }

    def __init__(self, argument_table, name, documentation, parse_type,
                 serialized_name):
        super(PageArgument, self).__init__(name)
        shape_resolver = model.ShapeResolver({})
        self.argument_model = shape_resolver.get_shape('PageArgument',
                                                       {'type': 'string'})
        self._serialized_name = serialized_name
        self._documentation = documentation
        self._parse_type = parse_type
        self._required = False
        self._shadowed_argument = None
        if name in argument_table:
            # If there's already an entry in the arg table for this argument,
            # this means we're shadowing an argument for this operation.  We
            # need to store this later in case pagination is turned off because
            # we put these arguments back.
            # See the comment in check_should_enable_pagination() for more info.
            self._shadowed_argument = argument_table[name]

    @property
    def cli_name(self):
        return '--' + self._name

    @property
    def cli_type(self):
        return self.cli_type_name

    @property
    def cli_type_name(self):
        return self._parse_type

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, value):
        self._required = value

    @property
    def documentation(self):
        return self._documentation

    @property
    def shadowed_argument(self):
        return self._shadowed_argument

    def add_to_parser(self, parser):
        parser.add_argument(self.cli_name, dest=self.py_name,
                            type=self.type_map[self._parse_type])

    def add_to_params(self, parameters, value):
        if value is not None:
            pagination_config = parameters.get('PaginationConfig', {})
            pagination_config[self._serialized_name] = value
            parameters['PaginationConfig'] = pagination_config



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import io
import os
import re
import sys
import threading
import time
import urllib.error
import urllib.parse
import urllib.request

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, CDP_PRIVATE_KEY_KEY_NAME
from cdpcli.exceptions import InteractiveLoginError
from cdpcli.extensions.configure import CREDENTIAL_FILE_COMMENT
from cdpcli.extensions.interactivelogin import LoginCommand
import mock
from tests import unittest
from tests.unit import FakeContext


class TestLoginCommand(unittest.TestCase):

    def setUp(self):
        self.writer = mock.Mock()
        self.open_new_browser = mock.Mock()
        self.context = FakeContext({'config_file': 'myconfigfile'})
        self.login = LoginCommand(config_writer=self.writer,
                                  open_new_browser=self.open_new_browser)

    def assert_login_url(self,
                         login_url,
                         expected_hostname,
                         expected_path=None,
                         expected_account_id=None,
                         expected_idp=None,
                         expected_return_url_port=None,
                         expected_extra_query=None):
        url_parsed = urllib.parse.urlsplit(login_url)
        url_params = urllib.parse.parse_qs(url_parsed.query)
        self.assertEqual(url_parsed.scheme, 'https')
        self.assertEqual(url_parsed.hostname, expected_hostname)
        self.assertEqual(url_parsed.path, expected_path)
        num_expected_queries = 2
        self.assertEqual(url_params.get('accountId'), [expected_account_id])
        if expected_idp is not None:
            num_expected_queries += 1
            self.assertEqual(url_params.get('idp'), [expected_idp])
        if expected_return_url_port is None:
            self.assertEqual(len(url_params.get('returnUrl')), 1)
            self.assertRegex(
                url_params.get('returnUrl')[0],
                'http://localhost:\\d+/interactiveLogin')
        else:
            self.assertEqual(len(url_params.get('returnUrl')), 1)
            self.assertEqual(
                url_params.get('returnUrl')[0],
                'http://localhost:%d/interactiveLogin' % expected_return_url_port)
        if expected_extra_query is not None:
            num_expected_queries += 1
            self.assertEqual(
                url_params.get(expected_extra_query.get('key'))[0],
                expected_extra_query.get('value'))
        self.assertEqual(len(url_params), num_expected_queries)

    def assert_credentials_file_updated_with(self,
                                             new_values,
                                             config_file_comment=None):
        called_args = self.writer.update_config.call_args_list
        credentials_file_call = called_args[0]
        expected_creds_file = os.path.expanduser('~/fake_credentials_filename')
        self.assertEqual(credentials_file_call,
                         mock.call(new_values,
                                   expected_creds_file,
                                   config_file_comment=config_file_comment))

    def mock_user_login(self, login_params, result, assert_func=None):
        i = 0
        while not self.open_new_browser.called:
            i += 1
            self.assertFalse(i > 20)
            time.sleep(0.1)
        url = self.open_new_browser.call_args[0][0]
        if assert_func is not None:
            assert_func(url)
        m = re.search('http%3A%2F%2Flocalhost%3A\\d+%2FinteractiveLogin', url)
        self.assertIsNotNone(m)
        return_url = urllib.parse.unquote(m.group())
        try:
            if login_params.startswith('/'):
                req_url = return_url.replace('/interactiveLogin', login_params)
            else:
                req_url = return_url + login_params
            resp = urllib.request.urlopen(req_url)
            result['status_code'] = resp.getcode()
            result['succeeded'] = True
        except urllib.error.HTTPError as e:
            result['status_code'] = e.code

    def test_login_command_open_browser(self):
        try:
            args = ['--account-id', 'test-account-guid', '--timeout', '1']
            parsed_globals = mock.Mock()
            parsed_globals.cdp_region = None
            self.login(self.context, args=args, parsed_globals=parsed_globals)
        except InteractiveLoginError:
            pass
        self.assertEqual(self.open_new_browser.call_count, 1)
        args, kwargs = self.open_new_browser.call_args
        self.assertEqual(len(args), 1)
        self.assert_login_url(login_url=args[0],
                              expected_hostname='consoleauth.altus.cloudera.com',
                              expected_path='/login',
                              expected_account_id='test-account-guid')

    def test_login_command_succeeded(self):
        login_result = {'succeeded': False}
        thread = threading.Thread(
            target=TestLoginCommand.mock_user_login,
            args=(self, '?accessKeyId=foo&privateKey=bar', login_result))
        thread.start()
        try:
            args = ['--account-id', 'foobar', '--timeout', '1']
            parsed_globals = mock.Mock()
            self.login(self.context, args=args, parsed_globals=parsed_globals)
            self.assert_credentials_file_updated_with(
                {CDP_ACCESS_KEY_ID_KEY_NAME: 'foo',
                 CDP_PRIVATE_KEY_KEY_NAME: 'bar'},
                config_file_comment=CREDENTIAL_FILE_COMMENT)
        finally:
            thread.join()
        self.assertEqual(login_result.get('status_code'), 200)
        self.assertTrue(login_result.get('succeeded'))

    def test_login_command_succeeded_for_profile(self):
        login_result = {'succeeded': False}
        thread = threading.Thread(
            target=TestLoginCommand.mock_user_login,
            args=(self, '?accessKeyId=foo&privateKey=bar', login_result))
        thread.start()
        try:
            args = ['--account-id', 'foobar', '--timeout', '1']
            parsed_globals = mock.Mock()
            parsed_globals.profile = 'myname'
            self.context.effective_profile = 'myname'
            self.login(self.context, args=args, parsed_globals=parsed_globals)
            self.assert_credentials_file_updated_with(
                {CDP_ACCESS_KEY_ID_KEY_NAME: 'foo',
                 CDP_PRIVATE_KEY_KEY_NAME: 'bar',
                 '__section__': 'myname'},
                config_file_comment=CREDENTIAL_FILE_COMMENT)
        finally:
            thread.join()
        self.assertEqual(login_result.get('status_code'), 200)
        self.assertTrue(login_result.get('succeeded'))

    def test_login_command_succeeded_for_profile_does_not_exist(self):
        login_result = {'succeeded': False}
        thread = threading.Thread(
            target=TestLoginCommand.mock_user_login,
            args=(self, '?accessKeyId=foo&privateKey=bar', login_result))
        thread.start()
        try:
            args = ['--account-id', 'foobar', '--timeout', '1']
            parsed_globals = mock.Mock()
            self.context.profile_does_not_exist = True
            self.login(self.context, args=args, parsed_globals=parsed_globals)
            self.assert_credentials_file_updated_with(
                {CDP_ACCESS_KEY_ID_KEY_NAME: 'foo',
                 CDP_PRIVATE_KEY_KEY_NAME: 'bar'},
                config_file_comment=CREDENTIAL_FILE_COMMENT)
        finally:
            thread.join()
        self.assertEqual(login_result.get('status_code'), 200)
        self.assertTrue(login_result.get('succeeded'))

    def test_login_command_succeeded_for_assigned_port(self):
        login_result = {'succeeded': False}
        thread = threading.Thread(
            target=TestLoginCommand.mock_user_login,
            args=(self, '?accessKeyId=foo&privateKey=bar', login_result))
        thread.start()
        try:
            args = ['--account-id', 'foobar', '--port', '10100', '--timeout', '1']
            parsed_globals = mock.Mock()
            self.login(self.context, args=args, parsed_globals=parsed_globals)
            args, kwargs = self.open_new_browser.call_args
            self.assertEqual(len(args), 1)
            self.assertTrue('10100' in args[0])
        finally:
            thread.join()
        self.assertEqual(login_result.get('status_code'), 200)
        self.assertTrue(login_result.get('succeeded'))

    def test_login_command_timeout(self):
        with self.assertRaisesRegexp(InteractiveLoginError,
                                     'Login timeout'):
            args = ['--account-id', 'foobar', '--timeout', '1']
            parsed_globals = mock.Mock()
            self.login(self.context, args=args, parsed_globals=parsed_globals)
        self.assertFalse(self.writer.update_config.called)

    def test_login_command_failed_missing_access_key_id(self):
        new_out = io.StringIO()
        old_out = sys.stderr
        try:
            sys.stderr = new_out
            login_result = {'succeeded': False}
            thread = threading.Thread(
                target=TestLoginCommand.mock_user_login,
                args=(self, '?privateKey=bar', login_result))
            thread.start()
            try:
                args = ['--account-id', 'foobar']
                parsed_globals = mock.Mock()
                self.login(self.context, args=args, parsed_globals=parsed_globals)
            except InteractiveLoginError:
                pass
            finally:
                thread.join()
            self.assertEqual(login_result.get('status_code'), 400)
            self.assertFalse(login_result.get('succeeded'))
            self.assertEqual(new_out.getvalue().strip(),
                             'Login failed: Missing access key id or private key')
        finally:
            sys.stderr = old_out

    def test_login_command_failed_missing_private_key(self):
        new_out = io.StringIO()
        old_out = sys.stderr
        try:
            sys.stderr = new_out
            login_result = {'succeeded': False}
            thread = threading.Thread(
                target=TestLoginCommand.mock_user_login,
                args=(self, '?accessKeyId=foo', login_result))
            thread.start()
            try:
                args = ['--account-id', 'foobar']
                parsed_globals = mock.Mock()
                self.login(self.context, args=args, parsed_globals=parsed_globals)
            except InteractiveLoginError:
                pass
            finally:
                thread.join()
            self.assertEqual(login_result.get('status_code'), 400)
            self.assertFalse(login_result.get('succeeded'))
            self.assertEqual(new_out.getvalue().strip(),
                             'Login failed: Missing access key id or private key')
        finally:
            sys.stderr = old_out

    def test_login_command_failed_not_interactive_login_path(self):
        new_out = io.StringIO()
        old_out = sys.stderr
        try:
            sys.stderr = new_out
            login_result = {'succeeded': False}
            thread = threading.Thread(
                target=TestLoginCommand.mock_user_login,
                args=(self, '/wrong_path', login_result))
            thread.start()
            try:
                args = ['--account-id', 'foobar']
                parsed_globals = mock.Mock()
                self.login(self.context, args=args, parsed_globals=parsed_globals)
            except InteractiveLoginError:
                pass
            finally:
                thread.join()
            self.assertEqual(login_result.get('status_code'), 404)
            self.assertFalse(login_result.get('succeeded'))
            self.assertEqual(new_out.getvalue().strip(),
                             'Login failed: URL path not supported')
        finally:
            sys.stderr = old_out

    def test_login_command_failed_error(self):
        new_out = io.StringIO()
        old_out = sys.stderr
        try:
            sys.stderr = new_out
            login_result = {'succeeded': False}
            thread = threading.Thread(
                target=TestLoginCommand.mock_user_login,
                args=(self, '?error=Error+Message', login_result))
            thread.start()
            args = ['--account-id', 'foobar']
            parsed_globals = mock.Mock()
            self.login(self.context, args=args, parsed_globals=parsed_globals)
            thread.join()
            self.assertEqual(login_result.get('status_code'), 200)
            self.assertEqual(new_out.getvalue().strip(), 'Login failed: Error Message')
        finally:
            sys.stderr = old_out

    def test_find_unused_port(self):
        port = LoginCommand._find_unused_port()
        self.assertTrue(port > 0)

    def test_account_id_and_idp_from_login_url(self):
        parsed_args = mock.Mock()
        parsed_args.account_id = 'test'
        parsed_args.identity_provider = 'test'
        parsed_args.login_url = 'https://unit.test/auth/?accountId=foo&idp=bar'
        parsed_args.no_save_token = False
        parsed_globals = mock.Mock()
        config = {
            'account_id': 'test',
            'identity_provider': 'test'
        }
        login_url = self.login._resolve_login_url(
            parsed_args, parsed_globals, config, 10101)
        self.assert_login_url(login_url=login_url,
                              expected_hostname='unit.test',
                              expected_path='/auth/',
                              expected_account_id='foo',
                              expected_idp='bar',
                              expected_return_url_port=10101)

    def test_login_url_from_input_parameter(self):
        parsed_args = mock.Mock()
        parsed_args.account_id = 'foo'
        parsed_args.identity_provider = 'bar'
        parsed_args.login_url = 'https://unit.test/auth/?id=abc'
        parsed_args.no_save_token = False
        parsed_globals = mock.Mock()
        config = {
            'account_id': 'test',
            'identity_provider': 'test',
            'login_url': 'http://test'
        }
        login_url = self.login._resolve_login_url(
            parsed_args, parsed_globals, config, 10101)
        self.assert_login_url(login_url=login_url,
                              expected_hostname='unit.test',
                              expected_path='/auth/',
                              expected_account_id='foo',
                              expected_idp='bar',
                              expected_return_url_port=10101,
                              expected_extra_query={'key': 'id', 'value': 'abc'})

    def test_login_url_from_config(self):
        parsed_args = mock.Mock()
        parsed_args.account_id = None
        parsed_args.identity_provider = None
        parsed_args.login_url = None
        parsed_args.no_save_token = False
        parsed_globals = mock.Mock()
        config = {
            'account_id': 'foo',
            'identity_provider': 'bar',
            'login_url': 'https://unit.test/auth/?id=abc'
        }
        login_url = self.login._resolve_login_url(
            parsed_args, parsed_globals, config, 10101)
        self.assert_login_url(login_url=login_url,
                              expected_hostname='unit.test',
                              expected_path='/auth/',
                              expected_account_id='foo',
                              expected_idp='bar',
                              expected_return_url_port=10101,
                              expected_extra_query={'key': 'id', 'value': 'abc'})



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, CDP_PRIVATE_KEY_KEY_NAME
from cdpcli.extensions.configure import CREDENTIAL_FILE_COMMENT
from cdpcli.extensions.logout import LogoutCommand
import mock
from tests import unittest
from tests.unit import FakeContext


class TestLogoutCommand(unittest.TestCase):

    def setUp(self):
        self.writer = mock.Mock()
        self.context = FakeContext({'config_file': 'myconfigfile'})
        self.args = []
        self.logout = LogoutCommand(config_writer=self.writer)

    def assert_credentials_file_updated_with(self,
                                             new_values,
                                             config_file_comment=None):
        called_args = self.writer.update_config.call_args_list
        credentials_file_call = called_args[0]
        expected_creds_file = os.path.expanduser('~/fake_credentials_filename')
        self.assertEqual(credentials_file_call,
                         mock.call(new_values,
                                   expected_creds_file,
                                   config_file_comment=config_file_comment))

    def test_logout_command_succeeded_default_profile(self):
        parsed_globals = mock.Mock()
        self.logout(self.context, args=self.args, parsed_globals=parsed_globals)
        self.assert_credentials_file_updated_with(
            {CDP_ACCESS_KEY_ID_KEY_NAME: '',
             CDP_PRIVATE_KEY_KEY_NAME: ''},
            config_file_comment=CREDENTIAL_FILE_COMMENT)

    def test_logout_command_succeeded_for_profile(self):
        parsed_globals = mock.Mock()
        parsed_globals.profile = 'myname'
        self.context.effective_profile = 'myname'
        self.logout(self.context, args=self.args, parsed_globals=parsed_globals)
        self.assert_credentials_file_updated_with(
            {CDP_ACCESS_KEY_ID_KEY_NAME: '',
             CDP_PRIVATE_KEY_KEY_NAME: '',
             '__section__': 'myname'},
            config_file_comment=CREDENTIAL_FILE_COMMENT)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, \
                   CDP_ACCESS_TOKEN_KEY_NAME, \
                   CDP_PRIVATE_KEY_KEY_NAME, \
                   CDP_REGION_KEY_NAME
from cdpcli.compat import compat_input
from cdpcli.endpoint import EndpointResolver
from cdpcli.exceptions import ProfileNotFound
from cdpcli.extensions.commands import BasicCommand
from cdpcli.extensions.configure import CREDENTIAL_FILE_COMMENT
from cdpcli.extensions.configure.get import ConfigureGetCommand
from cdpcli.extensions.configure.list import ConfigureListCommand
from cdpcli.extensions.configure.set import ConfigureSetCommand
from cdpcli.extensions.writer import ConfigFileWriter

from . import mask_value


class InteractivePrompter(object):

    def get_value(self, current_value, config_name, prompt_text=''):
        if config_name in (CDP_ACCESS_KEY_ID_KEY_NAME, CDP_PRIVATE_KEY_KEY_NAME):
            current_value = mask_value(current_value)
        interactive_long_input = False
        if config_name == CDP_PRIVATE_KEY_KEY_NAME:
            # See THUN-222 for context on why this is necessary
            interactive_long_input = True
        response = compat_input(
            "%s [%s]: " % (prompt_text, current_value),
            interactive_long_input)
        if not response:
            # If the user hits enter, we return a value of None
            # instead of an empty string.  That way we can determine
            # whether or not a value has changed.
            response = None
        return response


class ConfigureCommand(BasicCommand):
    NAME = 'configure'
    DESCRIPTION = BasicCommand.FROM_FILE()
    SYNOPSIS = ('cdp configure [--profile profile-name]')
    EXAMPLES = (
        'To create a new default configuration for CDP Public Cloud::\n'
        '\n'
        '    $ cdp configure\n'
        '    CDP Access Key ID [None]: accesskey\n'
        '    CDP Private Key [None]: privatekey\n'
        '    CDP Endpoint URL (blank for public cloud) [None]:\n'
    )
    SUBCOMMANDS = [
        {'name': 'list', 'command_class': ConfigureListCommand},
        {'name': 'get', 'command_class': ConfigureGetCommand},
        {'name': 'set', 'command_class': ConfigureSetCommand},
    ]

    # If you want to add new values to prompt, update this list here.
    VALUES_TO_PROMPT = [
        # (config_name, prompt_text)
        (CDP_ACCESS_KEY_ID_KEY_NAME, "CDP Access Key ID"),
        (CDP_PRIVATE_KEY_KEY_NAME, "CDP Private Key"),
        (CDP_REGION_KEY_NAME, "CDP Region"),
        (EndpointResolver.CDP_ENDPOINT_URL_KEY_NAME,
         "CDP Endpoint URL (blank for public cloud)")
    ]

    def __init__(self, prompter=None, config_writer=None):
        super(ConfigureCommand, self).__init__()
        if prompter is None:
            prompter = InteractivePrompter()
        self._prompter = prompter
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer

    def _run_main(self, client_creator, parsed_args, parsed_globals):
        # Called when invoked with no args "cdp configure"
        new_values = {}
        # This is the config from the config file scoped to a specific
        # profile.
        try:
            context = client_creator.context
            config = context.get_scoped_config()
        except ProfileNotFound:
            config = {}
        for config_name, prompt_text in self.VALUES_TO_PROMPT:
            current_value = config.get(config_name)
            new_value = self._prompter.get_value(current_value, config_name,
                                                 prompt_text)
            if new_value is not None and new_value != current_value:
                new_values[config_name] = new_value

        config_filename = os.path.expanduser(
            context.get_config_variable('config_file'))
        if new_values:
            self._write_out_creds_file_values(context,
                                              new_values,
                                              parsed_globals.profile)
            if parsed_globals.profile is not None:
                new_values['__section__'] = (
                    'profile %s' % parsed_globals.profile)
            self._config_writer.update_config(new_values, config_filename)

    def _write_out_creds_file_values(self, context, new_values, profile_name):
        # The access_key/private_key/access_token are now *always* written to the shared
        # credentials file (~/.cdp/credentials).
        credential_file_values = {}
        if CDP_ACCESS_KEY_ID_KEY_NAME in new_values:
            credential_file_values[CDP_ACCESS_KEY_ID_KEY_NAME] = new_values.pop(
                CDP_ACCESS_KEY_ID_KEY_NAME)
        if CDP_PRIVATE_KEY_KEY_NAME in new_values:
            credential_file_values[CDP_PRIVATE_KEY_KEY_NAME] = new_values.pop(
                CDP_PRIVATE_KEY_KEY_NAME)
        if CDP_ACCESS_TOKEN_KEY_NAME in new_values:
            credential_file_values[CDP_ACCESS_TOKEN_KEY_NAME] = new_values.pop(
                CDP_ACCESS_TOKEN_KEY_NAME)
        if credential_file_values:
            if profile_name is not None:
                credential_file_values['__section__'] = profile_name
            shared_credentials_filename = os.path.expanduser(
                context.get_config_variable('credentials_file'))
            self._config_writer.update_config(
                credential_file_values,
                shared_credentials_filename,
                config_file_comment=CREDENTIAL_FILE_COMMENT)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli.compat import compat_open
from cdpcli.compat import six
import requests


class ResourceLoadingError(Exception):
    pass


def get_paramfile(path):
    """Load parameter based on a resource URI.

    It is possible to pass parameters to operations by referring
    to files or URI's.  If such a reference is detected, this
    function attempts to retrieve the data from the file or URI
    and returns it.  If there are any errors or if the ``path``
    does not appear to refer to a file or URI, a ``None`` is
    returned.

    """
    data = None
    if isinstance(path, six.string_types):
        for prefix, function_spec in PREFIX_MAP.items():
            if path.startswith(prefix):
                function, kwargs = function_spec
                data = function(prefix, path, **kwargs)
    return data


def get_file(prefix, path, mode):
    file_path = os.path.expandvars(os.path.expanduser(path[len(prefix):]))
    try:
        with compat_open(file_path, mode) as f:
            return f.read()
    except UnicodeDecodeError:
        raise ResourceLoadingError(
            'Unable to load paramfile (%s), text contents could '
            'not be decoded.  If this is a binary file, please use the '
            'fileb:// prefix instead of the file:// prefix.' % file_path)
    except (OSError, IOError) as e:
        raise ResourceLoadingError('Unable to load paramfile %s: %s' % (
            path, e))


def get_uri(prefix, uri):
    try:
        r = requests.get(uri)
        if r.status_code == 200:
            return r.text
        else:
            raise ResourceLoadingError(
                "received non 200 status code of %s" % (
                    r.status_code))
    except Exception as e:
        raise ResourceLoadingError('Unable to retrieve %s: %s' % (uri, e))


PREFIX_MAP = {
    'file://': (get_file, {'mode': 'r'}),
    'fileb://': (get_file, {'mode': 'rb'}),
    'http://': (get_uri, {}),
    'https://': (get_uri, {}),
}


class ParamFileVisitor(object):
    """
    This visitor's visit method will walk the input params object of the input
    shape, visiting all fields and recursing through complex fields. Any string
    field encountered will get paramfile resolution unless it is marked in the
    model as being x-no-paramfile.
    """

    def visit(self, params, shape):
        return self._visit(params, shape, name='')

    def _visit(self, params, shape, name):
        return getattr(self, '_visit_%s' % shape.type_name)(params, shape, name)

    def _visit_object(self, params, shape, name):
        visited = dict()
        for param in params:
            if param not in shape.members:
                visited[param] = params[param]
                continue
            visited[param] = self._visit(params[param],
                                         shape.members[param],
                                         '%s.%s' % (name, param))
        return visited

    def _visit_map(self, params, shape, name):
        visited = dict()
        for param in params:
            # does not support visiting key, as it is supposed to be string
            visited[param] = self._visit(params[param],
                                         shape.value,
                                         '%s.%s' % (name, param))
        return visited

    def _visit_boolean(self, param, shape, name):
        return param

    def _visit_integer(self, param, shape, name):
        return param

    def _visit_float(self, param, shape, name):
        return param

    def _visit_double(self, param, shape, name):
        return param

    def _visit_string(self, param, shape, name):
        if shape.is_no_paramfile:
            return param
        override = get_paramfile(param)
        if override is not None:
            return override
        return param

    def _visit_array(self, param, shape, name):
        visited = []
        for i, item in enumerate(param):
            visited.append(self._visit(item, shape.member, '%s[%s]' % (name, i)))
        return visited

    def _visit_datetime(self, param, shape, name):
        return param

    def _visit_blob(self, param, shape, name):
        if shape.is_no_paramfile:
            return param
        override = get_paramfile(param)
        if override is not None:
            return override
        return param



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import signal

from cdpcli.help import ExecutableNotFoundError
from cdpcli.help import HelpCommand, WindowsHelpRenderer
from cdpcli.help import PosixHelpRenderer
import mock
from tests import FileCreator
from tests import skip_if_windows
from tests import unittest


class HelpSpyMixin(object):
    def __init__(self):
        self.exists_on_path = {}
        self.popen_calls = []
        self.mock_popen = mock.Mock()

    def _exists_on_path(self, name):
        return self.exists_on_path.get(name)

    def _popen(self, *args, **kwargs):
        self.popen_calls.append((args, kwargs))
        return self.mock_popen


class FakePosixHelpRenderer(HelpSpyMixin, PosixHelpRenderer):
    pass


class FakeWindowsHelpRenderer(HelpSpyMixin, WindowsHelpRenderer):
    def __init__(self):
        HelpSpyMixin.__init__(self)
        WindowsHelpRenderer.__init__(self)


class TestHelpPager(unittest.TestCase):

    def setUp(self):
        self.environ = {}
        self.environ_patch = mock.patch('os.environ', self.environ)
        self.environ_patch.start()
        self.renderer = PosixHelpRenderer()

    def tearDown(self):
        self.environ_patch.stop()

    def test_no_env_vars(self):
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         self.renderer.PAGER.split())

    def test_manpager(self):
        pager_cmd = 'foobar'
        os.environ['MANPAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         pager_cmd.split())

    def test_pager(self):
        pager_cmd = 'fiebaz'
        os.environ['PAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         pager_cmd.split())

    def test_both(self):
        os.environ['MANPAGER'] = 'foobar'
        os.environ['PAGER'] = 'fiebaz'
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         'foobar'.split())

    def test_manpager_with_args(self):
        pager_cmd = 'less -X'
        os.environ['MANPAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         pager_cmd.split())

    def test_pager_with_args(self):
        pager_cmd = 'less -X --clearscreen'
        os.environ['PAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         pager_cmd.split())

    def test_no_groff_exists(self):
        renderer = FakePosixHelpRenderer()
        renderer.exists_on_path['groff'] = False
        with self.assertRaisesRegexp(ExecutableNotFoundError,
                                     'Could not find executable named: groff'):
            renderer.render('foo')

    def test_shlex_split_for_pager_var(self):
        pager_cmd = '/bin/sh -c "col -bx | vim -c \'set ft=man\' -"'
        os.environ['PAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(),
                         ['/bin/sh', '-c', "col -bx | vim -c 'set ft=man' -"])

    def test_can_render_contents(self):
        renderer = FakePosixHelpRenderer()
        renderer.exists_on_path['groff'] = True
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('foo')
        self.assertEqual(renderer.popen_calls[-1][0], (['less', '-R'],))

    def test_can_page_output_on_windows(self):
        renderer = FakeWindowsHelpRenderer()
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('foo')
        self.assertEqual(renderer.popen_calls[-1][0], (['more'],))

    def test_render_environments_help(self):
        renderer = FakePosixHelpRenderer()
        renderer.exists_on_path['groff'] = True
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('\n\n'
                        '.. _cli:cdp environments:\n\n\n'
                        '************\n'
                        'environments\n'
                        '************\n'
                        '\n\n\n'
                        ':subtitle: Cloudera Environments Service\n\n\n\n'
                        ':version: 0.9.36\n\n\n\n'
                        '===========\n'
                        'Description\n'
                        '===========\n'
                        '\n'
                        'CDP Environments Service description.\n'
                        '\n'
                        '=====================\n'
                        'Available Subcommands\n'
                        '=====================\n'
                        '\n\n\n\n'
                        '* create-environment\n\n')
        args, kargs = renderer.mock_popen.communicate.call_args_list[0]
        self.assertIn(b'Cloudera Environments Service', kargs.get('input'))
        self.assertIn(b'Generated by docutils manpage writer', kargs.get('input'))

    def test_render_environments_help_on_windows(self):
        renderer = FakeWindowsHelpRenderer()
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('\n\n'
                        '.. _cli:cdp environments:\n\n\n'
                        '************\n'
                        'environments\n'
                        '************\n'
                        '\n\n\n'
                        ':subtitle: Cloudera Environments Service\n\n\n\n'
                        ':version: 0.9.36\n\n\n\n'
                        '===========\n'
                        'Description\n'
                        '===========\n'
                        '\n'
                        'CDP Environments Service description.\n'
                        '\n'
                        '=====================\n'
                        'Available Subcommands\n'
                        '=====================\n'
                        '\n\n\n\n'
                        '* create-environment\n\n')
        args, kargs = renderer.mock_popen.communicate.call_args
        self.assertEqual(kargs, {'input': b'\n'
                                          b'environments\n'
                                          b'^^^^^^^^^^^^\n\n\n'
                                          b'Description\n'
                                          b'***********\n\n'
                                          b'CDP Environments Service description.\n\n\n'
                                          b'Available Subcommands\n'
                                          b'*********************\n\n'
                                          b'* create-environment\n'})

    @skip_if_windows("Ctrl-C not valid on windows.")
    def test_can_handle_ctrl_c(self):
        class CtrlCRenderer(FakePosixHelpRenderer):
            def _popen(self, *args, **kwargs):
                if self._is_pager_call(args):
                    os.kill(os.getpid(), signal.SIGINT)
                return self.mock_popen

            def _is_pager_call(self, args):
                return 'less' in args[0]

        renderer = CtrlCRenderer()
        renderer.mock_popen.communicate.return_value = ('send to pager', '')
        renderer.exists_on_path['groff'] = True
        renderer.render('foo')
        last_call = renderer.mock_popen.communicate.call_args_list[-1]
        self.assertEqual(last_call, mock.call(input='send to pager'))


class TestHelpCommandBase(unittest.TestCase):
    def setUp(self):
        self.file_creator = FileCreator()

    def tearDown(self):
        self.file_creator.remove_all()


class TestHelpCommand(TestHelpCommandBase):
    """Test some of the deeper functionality of the HelpCommand

    We do this by subclassing from HelpCommand and ensure it is behaving
    as expected.
    """
    def setUp(self):
        super(TestHelpCommand, self).setUp()
        self.doc_generator_mock = mock.Mock()
        self.subcommand_mock = mock.Mock()
        self.renderer = mock.Mock()

        class SampleHelpCommand(HelpCommand):
            GeneratorClass = self.doc_generator_mock

            @property
            def subcommand_table(sample_help_cmd_self):
                return {'mycommand': self.subcommand_mock}

        self.cmd = SampleHelpCommand(None, None, None)
        self.cmd.renderer = self.renderer

    def test_subcommand_call(self):
        self.cmd(None, ['mycommand'], None)
        self.subcommand_mock.assert_called_with([], None)
        self.assertFalse(self.doc_generator_mock.called)

    def test_regular_call(self):
        self.cmd(None, [], None)
        self.assertFalse(self.subcommand_mock.called)
        self.doc_generator_mock.assert_called_with(self.cmd)
        self.assertTrue(self.renderer.render.called)

    def test_invalid_subcommand(self):
        with mock.patch('sys.stderr') as f:
            with self.assertRaises(SystemExit):
                self.cmd(None, ['no-exist-command'], None)
        error_message = ''.join(arg[0][0] for arg in f.write.call_args_list)
        self.assertIn('Invalid choice', error_message)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli.compat import json
from nose.tools import assert_equal
from nose.tools import assert_in
from tests import executing_in_container


def test_cli_options_match_service_model_validator():
    # We skip this test if we're in a container since the CLI data is not simple
    # to find. This is fine, since no additional coverage is offered by running
    # this on different OSes with different versions of python.
    if executing_in_container():
        return

    cli_data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 '../../cdpcli/data/cli.json')
    with open(cli_data_path, 'r') as fp:
        cli_data = json.load(fp)
    # The service model validator makes sure that no request object has a property
    # that collides with a top level CLI option since this confuses the CLI parser.
    # The validator does this using a hardcoded list of the top level CLI options,
    # and that list is duplicated here. The source of truth for the CLI options is
    # the cli.json file. We believe these options change infrequently, and so we
    # avoid actually hooking up some build dependency between the validator and the
    # CLI and instead have this unit test which is meant to catch any divergence.
    #
    # So... if you add a CLI option, and this test fails. Update the CLI_OPTIONS
    # list in: services/libs/protocols-yaml-parent/scripts/validate_service_models.py
    # and then update this list here.
    validated_cli_options = ['version',
                             'debug',
                             'no-verify-tls',
                             'ca-bundle',
                             'endpoint-url',
                             'form-factor',
                             'force-ipv4',
                             'access-token',
                             'output',
                             'color',
                             'cli-read-timeout',
                             'cli-connect-timeout',
                             'no-paginate',
                             'auth-config',
                             'profile',
                             'cdp-region',
                             'ensure-ascii']
    assert_equal(sorted(cli_data['options'].keys()), sorted(validated_cli_options))


def test_cdp_region():
    # We skip this test if we're in a container since the CLI data is not simple
    # to find. This is fine, since no additional coverage is offered by running
    # this on different OSes with different versions of python.
    if executing_in_container():
        return

    cli_data_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                 '../../cdpcli/data/cli.json')
    with open(cli_data_path, 'r') as fp:
        cli_data = json.load(fp)

    validated_cdp_regions = ['us-west-1',
                             'eu-1',
                             'ap-1']
    cdp_regions = cli_data['options']['cdp-region']['choices']
    assert_equal(sorted(cdp_regions), sorted(validated_cdp_regions))

    # Verify the same list of regions is in the configure help description.
    configure_doc_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                      '../../cdpcli/examples/configure/_description.rst')
    with open(configure_doc_path, 'r') as fp:
        configure_doc = fp.read()

    for cdp_region in validated_cdp_regions:
        assert_in(' * ' + cdp_region, configure_doc)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli import compat
from cdpcli import text
from cdpcli.compat import json
from cdpcli.paginate import PageIterator
from cdpcli.table import ColorizedStyler
from cdpcli.table import MultiTable
from cdpcli.table import Styler
from cdpcli.utils import json_encoder


def is_response_paginated(response):
    return isinstance(response, PageIterator)


class Formatter(object):
    def __init__(self, args):
        self._args = args

    def _get_default_stream(self):
        return compat.get_stdout_text_writer()

    def _flush_stream(self, stream):
        try:
            stream.flush()
        except IOError:
            pass


class FullyBufferedFormatter(Formatter):
    def __call__(self, command_name, response, stream=None):
        if stream is None:
            # Retrieve stdout on invocation instead of at import time
            # so that if anything wraps stdout we'll pick up those changes
            # (specifically colorama on windows wraps stdout).
            stream = self._get_default_stream()
        if is_response_paginated(response):
            response_data = response.build_full_result()
        else:
            response_data = response
        try:
            self._format_response(command_name, response_data, stream)
        except IOError:
            # If the reading end of our stdout stream has closed the file
            # we can just exit.
            pass
        finally:
            # flush is needed to avoid the "close failed in file object
            # destructor" in python2.x (see http://bugs.python.org/issue11380).
            self._flush_stream(stream)


class JSONFormatter(FullyBufferedFormatter):

    def _format_response(self, command_name, response, stream):
        # For operations that have no response body (e.g. s3 put-object)
        # the response will be an empty string.  We don't want to print
        # that out to the user but other "falsey" values like an empty
        # dictionary should be printed.
        if response != {}:
            ensure_ascii = self._args.ensure_ascii
            if ensure_ascii is None:
                ensure_ascii = False
            json.dump(response, stream, indent=4, default=json_encoder,
                      ensure_ascii=ensure_ascii)
            stream.write('\n')


class TableFormatter(FullyBufferedFormatter):
    """Pretty print a table from a given response.

    The table formatter is able to take any generic response
    and generate a pretty printed table.  It does this without
    using the output definition from the model.

    """
    def __init__(self, args, table=None):
        super(TableFormatter, self).__init__(args)
        if args.color == 'auto':
            self.table = MultiTable(initial_section=False,
                                    column_separator='|')
        elif args.color == 'off':
            styler = Styler()
            self.table = MultiTable(initial_section=False,
                                    column_separator='|', styler=styler)
        elif args.color == 'on':
            styler = ColorizedStyler()
            self.table = MultiTable(initial_section=False,
                                    column_separator='|', styler=styler)
        else:
            raise ValueError("Unknown color option: %s" % args.color)

    def _format_response(self, command_name, response, stream):
        if self._build_table(command_name, response):
            try:
                self.table.render(stream)
            except IOError:
                # If they're piping stdout to another process which exits before
                # we're done writing all of our output, we'll get an error about a
                # closed pipe which we can safely ignore.
                pass

    def _build_table(self, title, current, indent_level=0):
        if not current:
            return False
        if title is not None:
            self.table.new_section(title, indent_level=indent_level)
        if isinstance(current, list):
            if isinstance(current[0], dict):
                self._build_sub_table_from_list(current, indent_level, title)
            else:
                for item in current:
                    if self._scalar_type(item):
                        self.table.add_row([item])
                    elif all(self._scalar_type(el) for el in item):
                        self.table.add_row(item)
                    else:
                        self._build_table(title=None, current=item)
        if isinstance(current, dict):
            # Render a single row section with keys as header
            # and the row as the values, unless the value
            # is a list.
            self._build_sub_table_from_dict(current, indent_level)
        return True

    def _build_sub_table_from_dict(self, current, indent_level):
        # Render a single row section with keys as header
        # and the row as the values, unless the value
        # is a list.
        headers, more = self._group_scalar_keys(current)
        if len(headers) == 1:
            # Special casing if a dict has a single scalar key/value pair.
            self.table.add_row([headers[0], current[headers[0]]])
        elif headers:
            self.table.add_row_header(headers)
            self.table.add_row([current[k] for k in headers])
        for remaining in more:
            self._build_table(remaining, current[remaining],
                              indent_level=indent_level + 1)

    def _build_sub_table_from_list(self, current, indent_level, title):
        headers, more = self._group_scalar_keys_from_list(current)
        self.table.add_row_header(headers)
        first = True
        for element in current:
            if not first and more:
                self.table.new_section(title,
                                       indent_level=indent_level)
                self.table.add_row_header(headers)
            first = False
            # Use .get() to account for the fact that sometimes an element
            # may not have all the keys from the header.
            self.table.add_row([element.get(header, '') for header in headers])
            for remaining in more:
                # Some of the non scalar attributes may not necessarily
                # be in every single element of the list, so we need to
                # check this condition before recursing.
                if remaining in element:
                    self._build_table(remaining, element[remaining],
                                      indent_level=indent_level + 1)

    def _scalar_type(self, element):
        return not isinstance(element, (list, dict))

    def _group_scalar_keys_from_list(self, list_of_dicts):
        # We want to make sure we catch all the keys in the list of dicts.
        # Most of the time each list element has the same keys, but sometimes
        # a list element will have keys not defined in other elements.
        headers = set()
        more = set()
        for item in list_of_dicts:
            current_headers, current_more = self._group_scalar_keys(item)
            headers.update(current_headers)
            more.update(current_more)
        headers = list(sorted(headers))
        more = list(sorted(more))
        return headers, more

    def _group_scalar_keys(self, current):
        # Given a dict, separate the keys into those whose values are
        # scalar, and those whose values aren't.  Return two lists,
        # one is the scalar value keys, the second is the remaining keys.
        more = []
        headers = []
        for element in current:
            if self._scalar_type(current[element]):
                headers.append(element)
            else:
                more.append(element)
        headers.sort()
        more.sort()
        return headers, more


class TextFormatter(Formatter):

    def __call__(self, command_name, response, stream=None):
        if stream is None:
            stream = self._get_default_stream()
        try:
            if is_response_paginated(response):
                response_data = response.build_full_result()
            else:
                response_data = response
            self._format_response(response_data, stream)
        finally:
            self._flush_stream(stream)

    def _format_response(self, response, stream):
        text.format_text(response, stream)


def get_formatter(format_type, args):
    if format_type == 'json':
        return JSONFormatter(args)
    elif format_type == 'text':
        return TextFormatter(args)
    elif format_type == 'table':
        return TableFormatter(args)
    raise ValueError("Unknown output type: %s" % format_type)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class CdpCLIError(Exception):
    """
    The base exception class for CDP CLI exceptions.
    """
    fmt = 'An unspecified error occured'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs


class ValidationError(CdpCLIError):
    """
    An exception occurred validating parameters.
    """
    fmt = "Invalid value ('{value}') for param {param} of type {type_name}"


class ParamValidationError(CdpCLIError):
    fmt = 'Parameter validation failed:\n{report}'


class DataNotFoundError(CdpCLIError):
    """
    The data associated with a particular path could not be loaded.
    """
    fmt = 'Unable to load data for: {data_path}'


class ExecutableNotFoundError(CdpCLIError):
    """
    The executable was not found.
    """
    fmt = 'Could not find executable named: {executable_name}'


class OperationNotPageableError(CdpCLIError):
    fmt = 'Operation cannot be paginated: {operation_name}'


class ClientError(Exception):
    MSG_TEMPLATE = (
        'An error occurred: {error_message} ('
        'Status Code: {http_status_code}; '
        'Error Code: {error_code}; '
        'Service: {service_name}; '
        'Operation: {operation_name}; '
        'Request ID: {request_id};)')

    def __init__(self, error_response, operation_name, service_name,
                 http_status_code, request_id):
        msg = self.MSG_TEMPLATE.format(
            error_code=error_response['error'].get('code', 'Unknown'),
            error_message=error_response['error'].get('message', 'Unknown'),
            operation_name=operation_name,
            service_name=service_name,
            http_status_code=http_status_code,
            request_id=request_id)
        super(ClientError, self).__init__(msg)
        self.http_status_code = http_status_code
        self.response = error_response


class UnseekableStreamError(CdpCLIError):
    """
    Need to seek a stream, but stream does not support seeking.
    """
    fmt = ('Need to rewind the stream {stream_object}, but stream '
           'is not seekable.')


class EndpointConnectionError(CdpCLIError):
    fmt = (
        'Could not connect to the endpoint URL: "{endpoint_url}"')


class IncompleteReadError(CdpCLIError):
    """
    HTTP response did not return expected number of bytes.
    """
    fmt = ('{actual_bytes} read, but total bytes '
           'expected is {expected_bytes}.')


class PaginationError(CdpCLIError):
    fmt = 'Error during pagination: {message}'


class UnknownSignatureVersionError(CdpCLIError):
    """
    Requested Signature Version is not known.
    """
    fmt = 'Unknown Signature Version: {signature_version}.'


class UnsupportedSignatureVersionError(CdpCLIError):
    """
    Error when trying to access a method on a client that does not exist.
    """
    fmt = 'Signature version is not supported: {signature_version}'


class NoCredentialsError(CdpCLIError):
    """
    No credentials could be found.
    """
    fmt = 'Unable to locate CDP credentials: {err_msg}.'


class UnknownCredentialError(CdpCLIError):
    """
    Tried to insert before/after an unregistered credential type.
    """
    fmt = 'Credential named {name} not found.'


class MalformedCredentialsError(CdpCLIError):
    """
    Malformed credentials found.
    """
    fmt = 'Malformed credentials found in {provider}, missing: {cred_var}'


class PartialCredentialsError(CdpCLIError):
    """
    Only partial credentials were found.
    """
    fmt = 'Partial credentials found in {provider}, missing: {cred_var}'


class BaseEndpointResolverError(CdpCLIError):
    """
    Base error for endpoint resolving errors.

    Should never be raised directly, but clients can catch
    this exception if they want to generically handle any errors
    during the endpoint resolution process.

    """


class NoRegionError(BaseEndpointResolverError):
    """
    No region was specified.
    """
    fmt = 'You must specify a region.'


class ProfileNotFound(CdpCLIError):
    """
    The specified configuration profile was not found in the
    configuration file.

    """
    fmt = 'The config profile ({profile}) could not be found'


class ConfigNotFound(CdpCLIError):
    """
    The specified configuration file could not be found.
    """
    fmt = 'The specified config file ({path}) could not be found.'


class ConfigParseError(CdpCLIError):
    """
    The configuration file could not be parsed.
    """
    fmt = 'Unable to parse config file: {path}'


class InvalidConfiguredFormFactor(CdpCLIError):
    """
    The configured form factor is not a valid value.
    """
    fmt = ('The configured form factor {form_factor} is invalid. '
           'Valid values are: {valid_form_factors}')


class ClusterTerminatingError(CdpCLIError):

    """
    The cluster is terminating or has already terminated.
    """
    fmt = 'Cluster {cluster_name} is terminating.'


class ClusterStartingError(CdpCLIError):

    """
    The cluster is starting.
    """
    fmt = 'Cluster {cluster_name} is starting.'


class ClusterFailedError(CdpCLIError):

    """
    The cluster failed to start.
    """
    fmt = 'Cluster {cluster_name} failed to start.'


class ClusterStatusNotFound(CdpCLIError):

    """
    Unable to find cluster status.
    """
    fmt = 'Unable to find {cluster_name}\'s status.'


class ClusterEndpointNotFound(CdpCLIError):

    """
    Unable to find cluster's Cloudera Manager Endpoint.
    """
    fmt = 'Unable to find {cluster_name}\'s Cloudera Manager Endpoint.'


class SSHNotFoundError(CdpCLIError):

    """
    SSH or Putty not available.
    """
    fmt = 'SSH or Putty not available.'


class WrongPuttyKeyError(CdpCLIError):

    """
    A wrong key has been used with a compatible program.
    """
    fmt = 'Key file file format is incorrect. Putty expects a ppk file.'


class MissingArgumentError(CdpCLIError):

    """
    The following argument is required.
    """
    fmt = 'The following argument is required: {arg_name}.'


class InteractiveLoginError(CdpCLIError):

    """
    Login failed.
    """
    fmt = 'Login failed: {err_msg}.'


class WrongSvcFormFactorError(CdpCLIError):

    """
    The service does not work under the current form factor.
    """
    fmt = ('The command {service_name} is not available under the {form_factor}'
           ' CDP form factor. It is only available for these form factors: '
           '{service_form_factors}. Check that your profile configuration '
           'specifies the correct form factor, or that the endpoint URL in '
           'profile configuration or on the command line points to the correct '
           'control plane.')


class WrongOpFormFactorError(CdpCLIError):

    """
    The operation does not work under the current form factor.
    """
    fmt = ('The subcommand {operation_name} under the command {service_name} is '
           'not available under the {form_factor} CDP form factor. It is only '
           'available for these form factors: {operation_form_factors}. Check '
           'that your profile configuration or explicit endpoint URL points to '
           'the correct control plane.')


class WrongArgFormFactorError(CdpCLIError):

    """
    The argument does not work under the current form factor.
    """
    fmt = ('The argument {arg_name} is not available under the {form_factor} '
           'CDP form factor. It is only available for these form factors: '
           '{arg_form_factors}. Check that your profile configuration '
           'or explicit endpoint URL points to the correct control plane.')


class ExtensionImportError(CdpCLIError):

    """
    Failed to import CLI extension.
    """
    fmt = 'Failed to import CLI extension \'{ext_name}\': {err}.'


class WorkloadServiceDiscoveryError(CdpCLIError):

    """
    The workload service-discovery failed.
    """
    fmt = 'Workload service-discovery error: {err_msg}.'


class RedirectExtensionError(CdpCLIError):

    """
    The redirect extension failed.
    """
    fmt = 'The redirect extension failed: {err_msg}.'


class DfExtensionError(CdpCLIError):

    """
    The DF extension failed.
    """
    fmt = ('The DF extension failed: {err_msg} '
           'Service name: {service_name}, operation name: {operation_name}')



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import base64

from cdpcli import validate
from cdpcli.compat import json
from cdpcli.compat import OrderedDict
from cdpcli.compat import six


def create_serializer():
    serializer = Serializer()
    validator = validate.ParamValidator()
    return validate.ParamValidationDecorator(validator, serializer)


class Serializer(object):
    DEFAULT_ENCODING = 'utf-8'

    def _get_base64(self, value):
        # Returns the base64-encoded version of value, handling
        # both strings and bytes. The returned value is a string
        # via the default encoding.
        if isinstance(value, six.string_types):
            value = value.encode(self.DEFAULT_ENCODING)
        return base64.b64encode(value).strip().decode(
            self.DEFAULT_ENCODING)

    def serialize_to_request(self, parameters, operation_model):
        # Don't serialize any parameter with a None value.
        filtered_parameters = OrderedDict(
            (k, v) for k, v in parameters.items() if v is not None)

        serialized = {}
        serialized['method'] = operation_model.http['method']
        serialized['headers'] = {'Content-Type': 'application/json'}
        serialized['url_path'] = operation_model.http['requestUri']

        serialized_body = OrderedDict()
        if len(filtered_parameters) != 0:
            self._serialize(serialized_body,
                            filtered_parameters,
                            operation_model.input_shape)
        serialized['body'] = json.dumps(serialized_body).encode(self.DEFAULT_ENCODING)

        return serialized

    def _serialize(self, serialized, value, shape, key=None):
        serialize_method_name = '_serialize_type_%s' % shape.type_name
        method = getattr(self, serialize_method_name, self._default_serialize)
        method(serialized, value, shape, key)

    def _serialize_type_object(self, serialized, value, shape, key):
        if key is not None:
            # If a key is provided, this is a result of a recursive call, so we
            # need to add a new child dict as the value of the passed in dict.
            # Below we will add all the structure members to the new serialized
            # dictionary we just created.
            serialized[key] = OrderedDict()
            serialized = serialized[key]

        for member_key, member_value in value.items():
            member_shape = shape.members[member_key]
            self._serialize(serialized, member_value, member_shape, member_key)

    def _serialize_type_array(self, serialized, value, shape, key):
        array_obj = []
        serialized[key] = array_obj
        for array_item in value:
            wrapper = {}
            # JSON list serialization is the only case where we aren't setting
            # a key on a dict.  We handle this by using a __current__ key on a
            # wrapper dict to serialize each list item before appending it to
            # the serialized list.
            self._serialize(wrapper, array_item, shape.member, "__current__")
            array_obj.append(wrapper["__current__"])

    def _serialize_type_blob(self, serialized, value, shape, key):
        # Blob args must be base64 encoded.
        # If value type is string/text, it is already base64 encoded, verified
        # in validate.py.
        # Use isinstance(six.string_types) because a string could be either str
        # or unicode in python2.
        if isinstance(value, six.string_types):
            serialized[key] = value
        else:
            serialized[key] = self._get_base64(value)

    def _default_serialize(self, serialized, value, shape, key):
        serialized[key] = value



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from cdpcli.argprocess import ParamError
from cdpcli.extensions.arguments import OverrideRequiredArgsArgument
from cdpcli.paramfile import get_paramfile
from cdpcli.paramfile import ParamFileVisitor


def add_cli_input_json(operation_model, argument_table):
    cli_input_json_argument = CliInputJSONArgument(operation_model)
    cli_input_json_argument.add_to_arg_table(argument_table)


class CliInputJSONArgument(OverrideRequiredArgsArgument):
    """This argument inputs a JSON string as the entire input for a command.

    Ideally, the value to this argument should be a filled out JSON file
    generated by ``--generate-cli-skeleton``. The items in the JSON string
    will not clobber other arguments entered into the command line.
    """
    ARG_DATA = {
        'name': 'cli-input-json',
        'help_text': 'Performs service operation based on the JSON string '
                     'provided. The JSON string follows the format provided '
                     'by ``--generate-cli-skeleton``. If other arguments are '
                     'provided on the command line, the CLI values will override '
                     'the JSON-provided values.'
    }

    def __init__(self, operation_model):
        super(CliInputJSONArgument, self).__init__()
        self._operation_model = operation_model

    def invoke(self,
               client_creator,
               operation_model,
               parameters,
               parsed_args,
               parsed_globals):
        return self.add_to_call_parameters(parameters, parsed_args)

    def add_to_call_parameters(self, call_parameters, parsed_args):

        # Check if ``--cli-input-json`` was specified in the command line.
        input_json = getattr(parsed_args, 'cli_input_json', None)
        if input_json is not None:
            # Retrieve the JSON from the file if needed.
            retrieved_json = get_paramfile(input_json)
            # Nothing was retrieved from the file. So assume the argument
            # is already a JSON string.
            if retrieved_json is None:
                retrieved_json = input_json
            try:
                # Try to load the JSON string into a python dictionary
                input_data = json.loads(retrieved_json)
            except ValueError as e:
                raise ParamError(
                    self.name, "Invalid JSON: %s\nJSON received: %s"
                               % (e, retrieved_json))
            # We run the ParamFileVisitor over the input data to resolve any
            # paramfile references in it.
            input_data = ParamFileVisitor().visit(
                input_data, self._operation_model.input_shape)
            # Add the members from the input JSON to the call parameters.
            self._update_call_parameters(call_parameters, input_data)
        return True

    def _update_call_parameters(self, call_parameters, input_data):
        for input_key in input_data.keys():
            # Only add the values to ``call_parameters`` if not already
            # present.
            if input_key not in call_parameters:
                call_parameters[input_key] = input_data[input_key]



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli.compat import six
from cdpcli.model import ServiceModel
from cdpcli.model import ShapeResolver
from cdpcli.paramfile import get_paramfile
from cdpcli.paramfile import ParamFileVisitor
from cdpcli.paramfile import ResourceLoadingError
import mock
from tests import FileCreator, unittest
from tests import skip_if_windows
import yaml

MODEL_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'paramfile')


class TestParamFile(unittest.TestCase):
    def setUp(self):
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()

    def test_text_file(self):
        contents = 'This is a test'
        filename = self.files.create_file('foo', contents)
        prefixed_filename = 'file://' + filename
        data = get_paramfile(prefixed_filename)
        self.assertEqual(data, contents)
        self.assertIsInstance(data, six.string_types)

    def test_binary_file(self):
        contents = 'This is a test'
        filename = self.files.create_file('foo', contents)
        prefixed_filename = 'fileb://' + filename
        data = get_paramfile(prefixed_filename)
        self.assertEqual(data, b'This is a test')
        self.assertIsInstance(data, six.binary_type)

    @skip_if_windows('Binary content error only occurs '
                     'on non-Windows platforms.')
    def test_cannot_load_text_file(self):
        contents = b'\xbfX\xac\xbe'
        filename = self.files.create_file('foo', contents, mode='wb')
        prefixed_filename = 'file://' + filename
        with self.assertRaises(ResourceLoadingError):
            get_paramfile(prefixed_filename)

    def test_file_does_not_exist_raises_error(self):
        with self.assertRaises(ResourceLoadingError):
            get_paramfile('file://file/does/not/existsasdf.txt')

    def test_no_match_uris_returns_none(self):
        self.assertIsNone(get_paramfile('foobar://somewhere.bar'))

    def test_non_string_type_returns_none(self):
        self.assertIsNone(get_paramfile(100))


class TestHTTPBasedResourceLoading(unittest.TestCase):
    def setUp(self):
        self.requests_patch = mock.patch('cdpcli.paramfile.requests')
        self.requests_mock = self.requests_patch.start()
        self.response = mock.Mock(status_code=200)
        self.requests_mock.get.return_value = self.response

    def tearDown(self):
        self.requests_patch.stop()

    def test_resource_from_http(self):
        self.response.text = 'http contents'
        loaded = get_paramfile('http://foo.bar.baz')
        self.assertEqual(loaded, 'http contents')
        self.requests_mock.get.assert_called_with('http://foo.bar.baz')

    def test_resource_from_https(self):
        self.response.text = 'http contents'
        loaded = get_paramfile('https://foo.bar.baz')
        self.assertEqual(loaded, 'http contents')
        self.requests_mock.get.assert_called_with('https://foo.bar.baz')

    def test_non_200_raises_error(self):
        self.response.status_code = 500
        with self.assertRaisesRegexp(ResourceLoadingError, 'foo\\.bar\\.baz'):
            get_paramfile('https://foo.bar.baz')

    def test_connection_error_raises_error(self):
        self.requests_mock.get.side_effect = Exception("Connection error.")
        with self.assertRaisesRegexp(ResourceLoadingError, 'foo\\.bar\\.baz'):
            get_paramfile('https://foo.bar.baz')


class TestParamFileVisitor(unittest.TestCase):
    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')
        self.resolver = ShapeResolver(self.model['definitions'])
        self.files = FileCreator()

    def tearDown(self):
        self.files.remove_all()

    def test_visitor(self):
        contents = 'This is a test'
        filename = self.files.create_file('jobOne.hql', contents)
        # We have modified our test model to mark jobXml with x-no-paramfile.
        params = {'clusterName': u'foo',
                  'jobs': [{'hiveJob': {'script': 'file://' + filename,
                                        'jobXml': 'file://' + filename}}]}
        shape = self.resolver.get_shape_by_name(
            'submit-jobs-request', 'SubmitJobsRequest')
        visited = ParamFileVisitor().visit(params, shape)
        params['jobs'][0]['hiveJob']['script'] = contents
        self.assertEqual(params, visited)

    def test_ref_map_visitor(self):
        contents = 'This is a test'
        filename = self.files.create_file('jobOne.hql', contents)
        # We have modified our test model to mark jobXml with x-no-paramfile.
        params = {'jobs': {'job1': {'hiveJob': {'script': 'file://' + filename,
                                    'jobXml': 'file://' + filename}}}}
        shape = self.resolver.get_shape_by_name(
            'map-paramfile-test', 'RefMapParamFileTest')
        visited = ParamFileVisitor().visit(params, shape)
        params['jobs']['job1']['hiveJob']['script'] = contents
        self.assertEqual(params, visited)

    def test_explicit_map_visitor(self):
        contents = 'This is a test'
        filename = self.files.create_file('jobOne.hql', contents)
        # We have modified our test model to mark jobXml with x-no-paramfile.
        params = {'jobs': {'job1': {'script': 'file://' + filename,
                                    'jobXml': 'file://' + filename}}}
        shape = self.resolver.get_shape_by_name(
            'map-paramfile-test', 'ExplicitMapParamFileTest')
        visited = ParamFileVisitor().visit(params, shape)
        params['jobs']['job1']['script'] = contents
        self.assertEqual(params, visited)

    def test_blob_visitor(self):
        contents = b'This is a test'
        filename = self.files.create_file('jobOne.hql', contents, mode='wb')
        # We have modified our test model to mark jobXml with x-no-paramfile.
        params = {'jobs': {'job1': {'script': 'fileb://' + filename,
                                    'jobXml': 'fileb://' + filename}}}
        shape = self.resolver.get_shape_by_name(
            'blob-test', 'BlobParamFileTest')
        visited = ParamFileVisitor().visit(params, shape)
        params['jobs']['job1']['script'] = contents
        self.assertEqual(params, visited)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, \
                   CDP_PRIVATE_KEY_KEY_NAME, \
                   CDP_REGION_KEY_NAME
from cdpcli.compat import six
from cdpcli.extensions.configure.list import ConfigureListCommand
import mock
from tests import unittest
from tests.unit import FakeContext


class TestConfigureListCommand(unittest.TestCase):

    def test_configure_list_command_nothing_set(self):
        # Test the case where the user only wants to change a single_value.
        context = FakeContext(
            all_variables={'config_file': '/config/location'})
        context.full_config = {
            'profiles': {'default': {}}}
        stream = six.StringIO()
        self.configure_list = ConfigureListCommand(stream)
        self.configure_list(context, args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegexpMatches(rendered, 'profile\\s+<not set>')
        self.assertRegexpMatches(rendered, 'cdp_region\\s+<not set>')
        self.assertRegexpMatches(rendered, 'cdp_access_key_id\\s+<not set>')
        self.assertRegexpMatches(rendered, 'cdp_private_key\\s+<not set>')

    def test_configure_from_env(self):
        env_vars = {
            'profile': 'myprofilename'
        }
        context = FakeContext(
            all_variables={'config_file': '/config/location'},
            environment_vars=env_vars)
        context.context_var_map = {'profile': (None, "PROFILE_ENV_VAR")}
        context.full_config = {
            'profiles': {'default': {}}}
        stream = six.StringIO()
        self.configure_list = ConfigureListCommand(stream)
        self.configure_list(context, args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegexpMatches(
            rendered, 'profile\\s+myprofilename\\s+env\\s+PROFILE_ENV_VAR')

    def test_configure_from_config_file(self):
        config_file_vars = {
            # this is not a known configuration so this is ignored.
            'foo': 'bar',
            CDP_REGION_KEY_NAME: 'eu-1',
            CDP_ACCESS_KEY_ID_KEY_NAME: 'key_id',
            CDP_PRIVATE_KEY_KEY_NAME: 'mysecretkey'
        }
        context = FakeContext(
            all_variables={'config_file': '/config/location'},
            config_file_vars=config_file_vars)
        context.context_var_map = {'region': ()}
        context.full_config = {
            'profiles': {'default': {}}}
        stream = six.StringIO()
        self.configure_list = ConfigureListCommand(stream)
        self.configure_list(context, args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegexpMatches(
            rendered, 'profile\\s+<not set>\\s+None\\s+None')
        self.assertRegexpMatches(
            rendered, 'cdp_region\\s+eu-1\\s+config-file')
        self.assertRegexpMatches(
            rendered, 'cdp_access_key_id\\s+\\*+y_id\\s+config-file')
        self.assertRegexpMatches(
            rendered, 'cdp_private_key\\s+\\*+tkey\\s+config-file')

    def test_configure_from_multiple_sources(self):
        # Here the profile is from an env var, the
        # region is from the config file, and the credentials
        # are from an iam-role.
        env_vars = {
            'profile': 'myprofilename'
        }
        config_file_vars = {
            CDP_REGION_KEY_NAME: 'eu-1'
        }
        credentials = mock.Mock()
        credentials.access_key_id = 'access_key'
        credentials.private_key = 'private_key'
        credentials.access_token = None
        credentials.method = 'foobar'
        context = FakeContext(
            all_variables={'config_file': '/config/location'},
            environment_vars=env_vars,
            config_file_vars=config_file_vars,
            credentials=credentials)
        context.context_var_map = {
            'profile': ('profile', 'CDP_DEFAULT_PROFILE')}
        context.full_config = {
            'profiles': {'default': {}}}
        stream = six.StringIO()
        self.configure_list = ConfigureListCommand(stream)
        self.configure_list(context, args=[], parsed_globals=None)
        rendered = stream.getvalue()
        # The profile came from an env var.
        self.assertRegexpMatches(
            rendered, 'profile\\s+myprofilename\\s+env\\s+CDP_DEFAULT_PROFILE')
        # The cdp_region came from config file.
        self.assertRegexpMatches(
            rendered, 'cdp_region\\s+eu-1\\s+config-file')
        # The credentials came from 'foobar'.  Note how we're
        # also checking that the access_key/private_key are masked
        # with '*' chars except for the last 4 chars.
        self.assertRegexpMatches(
            rendered, r'cdp_access_key_id\s+\*+_key\s+foobar')
        self.assertRegexpMatches(
            rendered, r'cdp_private_key\s+\*+_key\s+foobar')

    def test_profile_set_in_context(self):
        config_file_vars = {
            # this is not a known configuration so this is ignored.
            'foo': 'bar',
            CDP_REGION_KEY_NAME: 'eu-1',
            CDP_ACCESS_KEY_ID_KEY_NAME: 'key_id',
            CDP_PRIVATE_KEY_KEY_NAME: 'mysecretkey'
        }
        context = FakeContext(
            all_variables={'config_file': '/config/location'},
            config_file_vars=config_file_vars)
        context.profile = 'dev'
        context.context_var_map = {}
        context.full_config = {
            'profiles': {'dev': {}}}
        stream = six.StringIO()
        self.configure_list = ConfigureListCommand(stream)
        self.configure_list(context, args=[], parsed_globals=None)
        rendered = stream.getvalue()
        self.assertRegexpMatches(
            rendered, 'profile\\s+dev\\s+manual\\s+--profile')
        self.assertRegexpMatches(
            rendered, 'cdp_region\\s+eu-1\\s+config-file')
        self.assertRegexpMatches(
            rendered, r'cdp_access_key_id\s+\*+y_id\s+config-file')
        self.assertRegexpMatches(
            rendered, r'cdp_private_key\s+\*+tkey\s+config-file')



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli.thirdparty.six.moves import html_parser


class DocStringParser(html_parser.HTMLParser):
    """
    A simple HTML parser.  Focused on converting the subset of HTML
    that appears in the documentation strings of the JSON models into
    simple ReST format.
    """

    def __init__(self, doc):
        html_parser.HTMLParser.__init__(self)
        self.doc = doc
        self.unhandled_tags = []

    def handle_starttag(self, tag, attrs):
        handler_name = 'start_%s' % tag
        if hasattr(self.doc.style, handler_name):
            getattr(self.doc.style, handler_name)(attrs)
        else:
            self.unhandled_tags.append(tag)

    def handle_endtag(self, tag):
        handler_name = 'end_%s' % tag
        if hasattr(self.doc.style, handler_name):
            getattr(self.doc.style, handler_name)()

    def handle_data(self, data):
        if data.isspace():
            data = ' '
        else:
            end_space = data[-1].isspace()
            words = data.split()
            words = self.doc.translate_words(words)
            data = ' '.join(words)
            if end_space:
                data += ' '
        self.doc.handle_data(data)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import platform
import socket
import sys

from cdpcli import LIST_TYPE
from cdpcli import RELEASE
from cdpcli import VERSION
from cdpcli import xform_name
from cdpcli.argparser import ArgTableArgParser
from cdpcli.argparser import MainArgParser
from cdpcli.argparser import ServiceArgParser
from cdpcli.argprocess import unpack_argument
from cdpcli.arguments import BooleanArgument
from cdpcli.arguments import CLIArgument
from cdpcli.arguments import CustomArgument
from cdpcli.arguments import ListArgument
from cdpcli.arguments import UnknownArgumentError
from cdpcli.clicommand import CLICommand
from cdpcli.client import ClientCreator
from cdpcli.client import Context
from cdpcli.compat import copy_kwargs
from cdpcli.compat import OrderedDict
from cdpcli.compat import six
from cdpcli.config import Config
from cdpcli.endpoint import EndpointCreator
from cdpcli.endpoint import EndpointResolver
from cdpcli.exceptions import ExtensionImportError
from cdpcli.exceptions import InvalidConfiguredFormFactor
from cdpcli.exceptions import ProfileNotFound
from cdpcli.exceptions import WrongOpFormFactorError
from cdpcli.exceptions import WrongSvcFormFactorError
from cdpcli.extensions.arguments import OverrideRequiredArgsArgument
from cdpcli.extensions.cliinputjson import add_cli_input_json
from cdpcli.extensions.configure.classify import ClassifyDeployment
from cdpcli.extensions.configure.classify import DeploymentType
from cdpcli.extensions.configure.configure import ConfigureCommand
from cdpcli.extensions.generatecliskeleton import add_generate_skeleton
from cdpcli.extensions.interactivelogin import LoginCommand
from cdpcli.extensions.logout import LogoutCommand
from cdpcli.extensions.paginate import add_pagination_params
from cdpcli.extensions.paginate import check_should_enable_pagination
from cdpcli.extensions.refdoc import RefdocCommand
from cdpcli.formatter import get_formatter
from cdpcli.help import OperationHelpCommand
from cdpcli.help import ProviderHelpCommand
from cdpcli.help import ServiceHelpCommand
from cdpcli.loader import Loader
from cdpcli.model import ServiceModel
from cdpcli.paramfile import ParamFileVisitor
from cdpcli.paramformfactor import ParamFormFactorVisitor
from cdpcli.parser import ResponseParserFactory
from cdpcli.retryhandler import create_retry_handler
from cdpcli.translate import build_retry_config
from cdpcli.utils import get_extension_registers
import urllib3.util.connection as urllib3_connection

LOG = logging.getLogger('cdpcli.clidriver')
ROOT_LOGGER = logging.getLogger('')
LOG_FORMAT = ('%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s')


def main():
    driver = CLIDriver()
    return driver.main()


class CLIDriver(object):

    def __init__(self):
        self._loader = Loader()
        self._endpoint_creator = EndpointCreator(EndpointResolver())
        self._user_agent_header = self._build_user_agent_header()
        self._response_parser_factory = ResponseParserFactory()
        self._cli_data = self._loader.load_json('cli.json')
        self._retryhandler = self._create_default_retryhandler()
        self._available_services = self._loader.list_available_services()
        self._command_table = self._build_command_table()
        self._argument_table = self._build_argument_table()
        self._context = Context()
        self._client_creator = ClientCreator(self._loader,
                                             self._context,
                                             self._endpoint_creator,
                                             self._user_agent_header,
                                             self._response_parser_factory,
                                             self._retryhandler)
        self._form_factor = None

    def main(self, args=None):
        if args is None:
            args = sys.argv[1:]
        parser = self._create_parser()
        command_table = self._get_command_table()
        if len(args) == 0 or (len(args) == 1 and args[0] == '--help'):
            args = ['help']
        parsed_args, remaining = parser.parse_known_args(args)
        try:
            self._form_factor = self._get_form_factor(parsed_args)
            self._handle_top_level_args(parsed_args)
            self._filter_command_table_for_form_factor()
            self._warn_for_old_python()
            self._warn_for_non_public_release()
            return command_table[parsed_args.command](
                self._client_creator, remaining, parsed_args)
        except Exception as e:
            LOG.debug("Exception caught in main()", exc_info=True)
            sys.stderr.write("\n")
            sys.stderr.write("%s\n" % six.text_type(e))
            return 255

    def _get_loader(self):
        return self._loader

    def _get_cli_data(self):
        return self._cli_data

    def _get_command_table(self):
        return self._command_table

    def _build_user_agent_header(self):
        return 'CDPCLI/%s Python/%s %s/%s' % (VERSION,
                                              platform.python_version(),
                                              platform.system(),
                                              platform.release())

    def _build_command_table(self):
        commands = OrderedDict()
        services = self._get_available_services()
        for service_name in services:
            commands[service_name] = ServiceCommand(self, service_name)
        ConfigureCommand.add_command(commands)
        LoginCommand.add_command(commands)
        LogoutCommand.add_command(commands)
        RefdocCommand.add_command(commands)
        commands = OrderedDict(sorted(commands.items()))
        return commands

    def _get_form_factor(self, parsed_args):
        if parsed_args.command == 'refdoc':
            # Do not filter out any command if it is to generate help documents.
            return None

        # Find the form factor based on:
        # 1. the form factor explicitly specified by --form-factor, or else
        # 2. the configured form factor, or else
        # 3. the explicit endpoint URL, or else
        # 4. the configured CDP endpoint URL.
        if parsed_args.form_factor:
            form_factor = parsed_args.form_factor
        else:
            try:
                form_factor = self._context.get_scoped_config().get('form_factor', None)
            except ProfileNotFound:
                form_factor = None
            valid_form_factors = [dt.value for dt in list(DeploymentType)]
            if form_factor and form_factor not in valid_form_factors:
                raise InvalidConfiguredFormFactor(
                    form_factor=form_factor,
                    valid_form_factors=valid_form_factors)
            if not form_factor:
                endpoint_url = parsed_args.endpoint_url
                if not endpoint_url:
                    try:
                        endpoint_url = self._context.get_scoped_config(). \
                            get(EndpointResolver.CDP_ENDPOINT_URL_KEY_NAME, None)
                    except ProfileNotFound:
                        endpoint_url = None
                form_factor = \
                    ClassifyDeployment(endpoint_url).get_deployment_type().value
        LOG.debug("Current form factor is {}".format(form_factor))
        return form_factor

    def _filter_command_table_for_form_factor(self):
        """
        Replaces services and operations in the command table that do not apply
        to the current form factor with stubs that error out when called.
        """
        form_factor = self._form_factor

        if form_factor is None:
            # Do not filter out any command if form factor is None.
            # For example: 'refdoc' command.
            return

        for command in list(self._command_table.keys()):
            try:
                # If the service does not apply to the current form factor,
                # filter it out.
                service_model = self._command_table[command].service_model
                service_form_factors = service_model.form_factors
                if form_factor not in service_form_factors:
                    self._command_table[command] = \
                        FilteredServiceCommand(self, command, form_factor,
                                               service_form_factors)
                else:
                    for operation_name in service_model.operation_names:
                        # If the operation does not apply to the current form
                        # factor, filter it out.
                        operation_model = service_model.operation_model(operation_name)
                        operation_form_factors = operation_model.form_factors
                        if not operation_form_factors:
                            operation_form_factors = service_form_factors
                        if form_factor not in operation_form_factors:
                            self._command_table[command]. \
                                filter_operation(operation_name, form_factor,
                                                 operation_form_factors)

            except AttributeError:
                # not a service model, so available in all form factors
                pass

    def _get_argument_table(self):
        return self._argument_table

    def _build_argument_table(self):
        argument_table = OrderedDict()
        cli_data = self._get_cli_data()
        cli_arguments = cli_data.get('options', None)
        for option in cli_arguments:
            option_params = copy_kwargs(cli_arguments[option])
            cli_argument = self._create_cli_argument(option, option_params)
            cli_argument.add_to_arg_table(argument_table)
        return argument_table

    def _get_available_services(self):
        return self._available_services

    def get_service_model(self, service_name):
        service_data = self._loader.load_service_data(service_name)
        service_data['paths'] = OrderedDict(sorted(service_data.get('paths', {}).items()))
        return ServiceModel(service_data, service_name=service_name)

    def get_form_factor(self):
        return self._form_factor

    def _create_help_command(self):
        cli_data = self._get_cli_data()

        # We filter service aliases out of the service list at the bottom of the
        # top level help.
        commands = OrderedDict()
        for service_name, command in self._get_command_table().items():
            if not self._loader.is_service_alias(service_name):
                commands[service_name] = command

        return ProviderHelpCommand(commands,
                                   self._get_argument_table(),
                                   cli_data.get('description', None),
                                   cli_data.get('synopsis', None),
                                   cli_data.get('help_usage', None))

    def _create_parser(self):
        command_table = self._get_command_table()
        command_table['help'] = self._create_help_command()
        cli_data = self._get_cli_data()
        parser = MainArgParser(
            command_table,
            VERSION,
            cli_data.get('description', None),
            self._get_argument_table())
        return parser

    def _create_cli_argument(self, option_name, option_params):
        return CustomArgument(
            option_name,
            help_text=option_params.get('help', ''),
            dest=option_params.get('dest'),
            default=option_params.get('default'),
            action=option_params.get('action'),
            required=option_params.get('required'),
            choices=option_params.get('choices'),
            cli_type_name=option_params.get('type'),
            hidden=option_params.get('hidden', False))

    def _handle_top_level_args(self, args):
        if args.profile:
            self._client_creator.context.set_config_variable('profile',
                                                             args.profile)
        if args.auth_config:
            self._client_creator.context.set_config_variable('auth_config',
                                                             args.auth_config)
        if args.debug:
            self._setup_logger(logging.DEBUG)
            LOG.debug("CLI version: %s", self._user_agent_header)
            LOG.debug("Arguments entered to CLI: %s", sys.argv[1:])
        else:
            self._setup_logger(logging.WARNING)

        if args.force_ipv4:
            # Based on SO /a/46972341
            LOG.debug("Forcing IPv4 connections only")

            def _allowed_gai_family():
                return socket.AF_INET
            urllib3_connection.allowed_gai_family = _allowed_gai_family

    def _setup_logger(self, log_level):
        ROOT_LOGGER.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        formatter = logging.Formatter(LOG_FORMAT)
        ch.setFormatter(formatter)
        ROOT_LOGGER.addHandler(ch)

    def _create_default_retryhandler(self):
        # We create one retryhandler based on the __default__ configuration in
        # the _retry.json (in the 'data' directory). This retryhandler is used
        # by all services.
        config = self._load_retry_config()
        if not config:
            return
        LOG.info("Using retry config: %s" % config)
        return create_retry_handler(config)

    def _load_retry_config(self):
        original_config = self._loader.load_json('_retry.json')
        retry_config = build_retry_config(
            original_config['retry'],
            original_config.get('definitions', {}))
        return retry_config

    def _warn_for_old_python(self):
        if sys.version_info[0] < 3 or \
                (sys.version_info[0] == 3 and sys.version_info[1] < 6):
            LOG.warn('You are running the CDP CLI under Python %s. The CDP CLI '
                     'now requires Python 3.6 or higher. Please upgrade now to '
                     'avoid CLI errors.', sys.version)

    def _warn_for_non_public_release(self):
        if RELEASE != 'PUBLIC':
            if RELEASE == 'INTERNAL':
                article = 'an'
            else:
                article = 'a'
            LOG.warn('You are running {0} {1} release of the CDP CLI, which '
                     'has different capabilities from the standard public '
                     'release. Find the public release at: '
                     'https://pypi.org/project/cdpcli/'.format(article, RELEASE))


class ServiceCommand(CLICommand):
    """
    A top-level CLI command, corresponding to an API service.
    """

    def __init__(self, clidriver, name):
        self._clidriver = clidriver
        self._name = name
        self._command_table = None
        self._lineage = [self]
        self._service_model = None

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def service_model(self):
        return self._get_service_model()

    @property
    def lineage(self):
        return self._lineage

    @lineage.setter
    def lineage(self, value):
        self._lineage = value

    def _get_command_table(self):
        if self._command_table is None:
            self._command_table = self._create_command_table()
        return self._command_table

    def _get_service_model(self):
        if self._service_model is None:
            self._service_model = self._clidriver.get_service_model(self._name)
        return self._service_model

    def __call__(self, client_creator, args, parsed_globals):
        # Once we know we're trying to call a service for this operation
        # we can go ahead and create the parser for it.  We
        # can also grab the Service object from botocore.
        service_parser = self._create_parser()
        parsed_args, remaining = service_parser.parse_known_args(args)
        command_table = self._get_command_table()
        return command_table[parsed_args.operation](
            client_creator, remaining, parsed_globals)

    def _create_command_table(self):
        command_table = OrderedDict()
        service_model = self._get_service_model()
        for operation_name in service_model.operation_names:
            cli_name = xform_name(operation_name, '-')
            operation_model = service_model.operation_model(operation_name)
            command_table[cli_name] = ServiceOperation(
                clidriver=self._clidriver,
                name=cli_name,
                parent_name=self._name,
                operation_model=operation_model,
                operation_caller=CLIOperationCaller())
        register_ext, register_cmd = get_extension_registers(self._name)
        if register_cmd is not None:
            register_cmd(self._clidriver, service_model, command_table)
        self._add_lineage(command_table)
        return command_table

    def filter_operation(self, operation_name, form_factor, operation_form_factors):
        """
        Replace the named operation in this command's command table with a
        filtered one.
        """
        command_table = self._get_command_table()
        cli_name = xform_name(operation_name, '-')
        command_table[cli_name] = FilteredServiceOperation(
            clidriver=self._clidriver,
            name=cli_name,
            parent_name=self._name,
            form_factor=form_factor,
            operation_form_factors=operation_form_factors)

    def _add_lineage(self, command_table):
        for command in command_table:
            command_obj = command_table[command]
            command_obj.lineage = self.lineage + [command_obj]

    def create_help_command(self):
        command_table = OrderedDict()
        for command_name, command in self._get_command_table().items():
            command_table[command_name] = command
        return ServiceHelpCommand(obj=self._get_service_model(),
                                  command_table=command_table,
                                  arg_table=None,
                                  command_lineage='.'.join(self.lineage_names),
                                  name=self._name)

    def _create_parser(self):
        command_table = self._get_command_table()
        # Also add a 'help' command.
        command_table['help'] = self.create_help_command()
        return ServiceArgParser(
            operations_table=command_table, service_name=self._name)


class FilteredServiceCommand(ServiceCommand):
    """
    A stub service command that fails when run due to being under the wrong
    CLI form factor.
    """

    def __init__(self, clidriver, name, form_factor, service_form_factors):
        super().__init__(clidriver, name)
        self._clidriver = clidriver
        self._name = name
        self._form_factor = form_factor
        self._service_form_factors = service_form_factors

    def __call__(self, client_creator, args, parsed_globals):
        raise WrongSvcFormFactorError(
            service_name=self._name,
            form_factor=self._form_factor,
            service_form_factors=', '.join(self._service_form_factors))


class ServiceOperation(object):
    ARG_TYPES = {
        LIST_TYPE: ListArgument,
        'boolean': BooleanArgument,
    }
    DEFAULT_ARG_CLASS = CLIArgument

    def __init__(self, clidriver, name, parent_name, operation_caller, operation_model):
        self._clidriver = clidriver
        self._arg_table = None
        self._name = name
        # These is used so we can figure out what the proper event
        # name should be <parent name>.<name>.
        self._parent_name = parent_name
        # We can have more than one operation callers. They are called in
        # order and if any returns 'False' no other callers will be called.
        self._operation_callers = [operation_caller]
        self._lineage = [self]
        self._operation_model = operation_model

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def lineage(self):
        return self._lineage

    @lineage.setter
    def lineage(self, value):
        self._lineage = value

    @property
    def lineage_names(self):
        # Represents the lineage of a command in terms of command ``name``
        return [cmd.name for cmd in self.lineage]

    @property
    def arg_table(self):
        if self._arg_table is None:
            self._arg_table = self._create_argument_table()
        return self._arg_table

    def __call__(self, client_creator, args, parsed_globals):
        # Handle extensions first, so OverrideRequiredArgs (CliInputJson,
        # GenerateCliSkeleton, etc) could have a chance to run.
        self._handle_extensions()
        # We need to handle overriding required arguments before we create
        # the parser as the parser will parse the arguments and decide which
        # argument is required before we have a chance to modify the argument
        # table.
        self._handle_override_required_args(args)
        # Once we know we're trying to call a particular operation
        # of a service we can go ahead and load the parameters.
        operation_parser = self._create_operation_parser(self.arg_table)
        self._add_help(operation_parser)
        parsed_args, remaining = operation_parser.parse_known_args(args)
        if parsed_args.help == 'help':
            return self.create_help_command()(
                client_creator, remaining, parsed_globals)
        elif parsed_args.help:
            remaining.append(parsed_args.help)
        if remaining:
            raise UnknownArgumentError(
                "Unknown options: %s" % ', '.join(remaining))
        check_should_enable_pagination(self._arg_table,
                                       self._operation_model,
                                       parsed_args,
                                       parsed_globals)
        call_parameters = self._build_call_parameters(parsed_args,
                                                      self.arg_table,
                                                      self._clidriver.get_form_factor())
        return self._invoke_operation_callers(client_creator,
                                              call_parameters,
                                              parsed_args,
                                              parsed_globals)

    def create_help_command(self):
        return OperationHelpCommand(
            operation_model=self._operation_model,
            arg_table=self.arg_table,
            name=self._name,
            command_lineage='.'.join(self.lineage_names))

    def _add_help(self, parser):
        # The 'help' output is processed a little differently from
        # the operation help because the arg_table has
        # CLIArguments for values.
        parser.add_argument('help', nargs='?')

    def _build_call_parameters(self, args, arg_table, form_factor):
        # We need to convert the args specified on the command
        # line as valid **kwargs we can hand to botocore.
        service_params = {}
        # args is an argparse.Namespace object so we're using vars()
        # so we can iterate over the parsed key/values.
        parsed_args = vars(args)
        for arg_object in arg_table.values():
            py_name = arg_object.py_name
            if py_name in parsed_args:
                value = parsed_args[py_name]
                value = unpack_argument(arg_object, value)
                arg_object.add_to_params(service_params, value)
        # We run the ParamFormFactorVisitor over the input data to check
        # the form factor for arguments.
        ParamFormFactorVisitor(form_factor).visit(
            service_params, self._operation_model.input_shape)
        # We run the ParamFileVisitor over the input data to resolve any
        # paramfile references in it.
        service_params = ParamFileVisitor().visit(
            service_params, self._operation_model.input_shape)
        return service_params

    def _create_argument_table(self):
        argument_table = OrderedDict()
        input_shape = self._operation_model.input_shape
        required_arguments = []
        arg_dict = {}
        if input_shape is not None:
            required_arguments = input_shape.required_members
            arg_dict = input_shape.members
        for arg_name, arg_shape in arg_dict.items():
            cli_arg_name = xform_name(arg_name, '-')
            arg_class = self.ARG_TYPES.get(arg_shape.type_name,
                                           self.DEFAULT_ARG_CLASS)
            is_required = arg_name in required_arguments
            arg_object = arg_class(
                name=cli_arg_name,
                argument_model=arg_shape,
                is_required=is_required,
                operation_model=self._operation_model,
                serialized_name=arg_name,
                no_paramfile=arg_shape.is_no_paramfile)
            arg_object.add_to_arg_table(argument_table)
        add_pagination_params(self._operation_model, argument_table)
        add_cli_input_json(self._operation_model, argument_table)
        add_generate_skeleton(self._operation_model, argument_table)
        return argument_table

    def _create_operation_parser(self, arg_table):
        return ArgTableArgParser(arg_table, service_name=self._parent_name,
                                 operation_name=self._name)

    def _handle_override_required_args(self, args):
        argument_table = self.arg_table
        for cli_name, cli_argument in argument_table.items():
            if isinstance(cli_argument, OverrideRequiredArgsArgument):
                cli_argument.override_required_args(argument_table, args)
                self._operation_callers.insert(0, cli_argument)

    def _handle_extensions(self):
        if self._operation_model.extensions:
            # Iterate in reversed order to keep the execution order:
            # First extension should run first.
            for ext_name in reversed(self._operation_model.extensions):
                register_ext, register_cmd = get_extension_registers(ext_name)
                if register_ext is None:
                    raise ExtensionImportError(ext_name=ext_name, err='Not Found')
                register_ext(self._operation_callers, self._operation_model)

    def _invoke_operation_callers(self,
                                  client_creator,
                                  call_parameters,
                                  parsed_args,
                                  parsed_globals):
        def _create_client(service_name):
            # The TLS verification value can be a boolean or a CA_BUNDLE path. This
            # is a little odd, but ultimately comes from the python HTTP requests
            # library we're using.
            tls_verification = parsed_globals.verify_tls
            ca_bundle = getattr(parsed_globals, 'ca_bundle', None)
            if parsed_globals.verify_tls and ca_bundle is not None:
                tls_verification = ca_bundle

            # Retrieve values passed for extra client configuration.
            config_kwargs = {}
            if parsed_globals.read_timeout is not None:
                config_kwargs['read_timeout'] = int(parsed_globals.read_timeout)
            if parsed_globals.connect_timeout is not None:
                config_kwargs['connect_timeout'] = int(parsed_globals.connect_timeout)
            config = Config(**config_kwargs)

            client = client_creator.create_client(
                service_name,
                parsed_globals.endpoint_url,
                parsed_globals.cdp_region,
                tls_verification,
                client_creator.context.get_credentials(parsed_globals),
                client_config=config)
            return client

        for operation_caller in self._operation_callers:
            # Create a new client for each operation_caller because parsed_args and
            # parsed_globals could be changed in each iteration.
            if operation_caller.invoke(
                    _create_client,
                    self._operation_model,
                    call_parameters,
                    parsed_args,
                    parsed_globals) is False:
                break
        return 0


class FilteredServiceOperation(ServiceOperation):
    """
    A stub service operation that fails when run due to being under the wrong
    CLI form factor.
    """

    def __init__(self, clidriver, name, parent_name, form_factor, operation_form_factors):
        super().__init__(clidriver, name, parent_name,
                         operation_caller=None, operation_model=None)
        self._form_factor = form_factor
        self._operation_form_factors = operation_form_factors

    def __call__(self, client_creator, args, parsed_globals):
        raise WrongOpFormFactorError(
            operation_name=self._name,
            service_name=self._parent_name,
            form_factor=self._form_factor,
            operation_form_factors=', '.join(self._operation_form_factors))


class CLIOperationCaller(object):

    def invoke(self,
               client_creator,
               operation_model,
               parameters,
               parsed_args,
               parsed_globals):
        service_name = operation_model.service_model.service_name
        operation_name = operation_model.name
        client = client_creator(service_name)
        py_operation_name = xform_name(operation_name)
        if client.can_paginate(py_operation_name) and parsed_globals.paginate:
            response = client.get_paginator(
                py_operation_name).paginate(**parameters)
        else:
            response = getattr(client, xform_name(operation_name))(**parameters)
        self._display_response(operation_name, response, parsed_globals)
        return True

    def _display_response(self, command_name, response, parsed_globals):
        output = parsed_globals.output
        if output is None:
            output = "json"
        formatter = get_formatter(output, parsed_globals)
        formatter(command_name, response)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import shutil
import tempfile

from cdpcli.argprocess import ParamError
from cdpcli.extensions.cliinputjson import add_cli_input_json
from cdpcli.extensions.cliinputjson import CliInputJSONArgument
from cdpcli.model import DenormalizedStructureBuilder
import mock
from tests import unittest


def _make_file_uri(file_name):
    # this is needed for windows but
    # just to be sure on mac/unix
    file_name = file_name.replace("\\", "/")
    return file_name


class TestCliInputJSONArgument(unittest.TestCase):
    def setUp(self):
        # Make an arbitrary input model shape.
        self.input_shape = {
            'A': {'type': 'string'},
            'B': {'type': 'string'}
        }
        shape = DenormalizedStructureBuilder().with_members(
            self.input_shape).build_model()
        self.operation_model = mock.Mock(input_shape=shape)
        self.argument = CliInputJSONArgument(self.operation_model)

        # Create the various forms the data could come in. The two main forms
        # are as a string and or as a path to a file.
        self.input_json = '{"A": "foo", "B": "bar"}'

        # Make a temporary file
        self.temp_dir = tempfile.mkdtemp()
        self.temp_file = os.path.join(self.temp_dir, 'foo.json')
        with open(self.temp_file, 'w') as f:
            f.write(self.input_json)

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_add_argument_action(self):
        argument_table = {}
        add_cli_input_json(self.operation_model, argument_table)
        arg_handler = argument_table['cli-input-json']
        self.assertIsNotNone(arg_handler)
        self.assertTrue(isinstance(arg_handler, CliInputJSONArgument))

    def test_add_to_call_parameters_no_file(self):
        parsed_args = mock.Mock()
        # Make the value a JSON string
        parsed_args.cli_input_json = self.input_json
        call_parameters = {}
        self.argument.add_to_call_parameters(
            call_parameters=call_parameters,
            parsed_args=parsed_args
        )
        self.assertEqual(call_parameters, {'A': 'foo', 'B': 'bar'})

    def test_add_to_call_parameters_with_file(self):
        parsed_args = mock.Mock()
        # Make the value a file with JSON located inside.
        parsed_args.cli_input_json = 'file://' + self.temp_file
        call_parameters = {}
        self.argument.add_to_call_parameters(
            call_parameters=call_parameters,
            parsed_args=parsed_args
        )
        self.assertEqual(call_parameters, {'A': 'foo', 'B': 'bar'})

    def test_add_to_call_parameters_with_file_from_visitor(self):
        temp_file1 = os.path.join(self.temp_dir, 'foo.json')
        with open(temp_file1, 'w') as f:
            f.write("baz")
        temp_file1 = _make_file_uri(temp_file1)
        temp_file2 = os.path.join(self.temp_dir, 'bar.json')
        with open(temp_file2, 'w') as f:
            f.write('{"A": "foo", "B": "file://%s"}' % temp_file1)
        temp_file2 = _make_file_uri(temp_file2)
        parsed_args = mock.Mock()
        parsed_args.cli_input_json = 'file://' + temp_file2
        call_parameters = {}
        self.argument.add_to_call_parameters(
            call_parameters=call_parameters,
            parsed_args=parsed_args
        )
        self.assertEqual(call_parameters, {'A': 'foo', 'B': 'baz'})

    def test_add_to_call_parameters_bad_json(self):
        parsed_args = mock.Mock()
        # Create a bad JSON input
        parsed_args.cli_input_json = self.input_json + ','
        call_parameters = {}
        with self.assertRaises(ParamError):
            self.argument.add_to_call_parameters(
                call_parameters=call_parameters,
                parsed_args=parsed_args
            )

    def test_add_to_call_parameters_no_clobber(self):
        parsed_args = mock.Mock()
        parsed_args.cli_input_json = self.input_json
        # The value for ``A`` should not be clobbered by the input JSON
        call_parameters = {'A': 'baz'}
        self.argument.add_to_call_parameters(
            call_parameters=call_parameters,
            parsed_args=parsed_args
        )
        self.assertEqual(call_parameters, {'A': 'baz', 'B': 'bar'})

    def test_no_add_to_call_parameters(self):
        parsed_args = mock.Mock()
        parsed_args.cli_input_json = None
        call_parameters = {'A': 'baz'}
        self.argument.add_to_call_parameters(
            call_parameters=call_parameters,
            parsed_args=parsed_args
        )
        # Nothing should have been added to the call parameters because
        # ``cli_input_json`` is not in the ``parsed_args``
        self.assertEqual(call_parameters, {'A': 'baz'})



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from base64 import b64decode, urlsafe_b64encode
from email.utils import formatdate
import hashlib
import logging

from asn1crypto import keys, pem
from cdpcli.compat import json
from cdpcli.compat import OrderedDict
from cdpcli.compat import urlsplit
from cdpcli.exceptions import NoCredentialsError
from ecdsa import SigningKey
from ecdsa.util import sigencode_der
from pure25519 import eddsa
import rsa


LOG = logging.getLogger('cdpcli.auth')


class BaseSigner(object):
    def add_auth(self, request):
        raise NotImplementedError("add_auth")


class V1Signer(object):
    ERROR_MESSAGE = \
       "Failed to import private key from: '%s'. The private key is " \
       "corrupted or not in the right format. The private key " \
       "was extracted either from 'env' (environment variables), " \
       "'shared-credentials-file' (a profile in the shared " \
       "credential file, by default under ~/.cdp/credentials), or " \
       "'auth-config-file' (a file containing the credentials whose " \
       "location was supplied on the command line.)"

    def __init__(self, credentials, auth_method):
        self.credentials = credentials
        self.auth_method = auth_method

    def _raw_sign_string(self, string_to_sign):
        raise NotImplementedError("Implement _raw_sign_string")

    def _sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the base64
        encoded signature in UTF-8 format.
        :param string_to_sign: String to sign
        :return: Signature as string
        """
        signature = self._raw_sign_string(string_to_sign)
        return urlsafe_b64encode(signature).strip().decode('utf-8')

    def _canonical_standard_headers(self, headers):
        interesting_headers = ['content-type', 'x-altus-date']
        hoi = []
        if 'x-altus-date' in headers:
            raise Exception("x-altus-date found in headers!")
        headers['x-altus-date'] = self._get_date()
        for ih in interesting_headers:
            found = False
            for key in headers:
                lk = key.lower()
                if headers[key] is not None and lk == ih:
                    hoi.append(headers[key].strip())
                    found = True
            if not found:
                hoi.append('')
        return '\n'.join(hoi)

    def _canonical_string(self, method, split, headers):
        cs = method.upper() + '\n'
        cs += self._canonical_standard_headers(headers) + '\n'
        cs += split.path + '\n'
        cs += self.auth_method
        return cs

    def _get_signature(self, method, split, headers):
        string_to_sign = self._canonical_string(method, split, headers)
        LOG.debug('StringToSign:\n%s', string_to_sign)
        return self._sign_string(string_to_sign)

    def add_auth(self, request):
        if self.credentials is None:
            raise NoCredentialsError(
                err_msg='Credentials not available for request signing')
        LOG.debug("Calculating signature using %s." % self.auth_method)
        LOG.debug('HTTP request method: %s', request.method)
        split = urlsplit(request.url)
        signature = self._get_signature(request.method,
                                        split,
                                        request.headers)
        self._inject_signature(request, signature)

    def _get_date(self):
        return formatdate(usegmt=True)

    def _inject_signature(self, request, signature):
        if 'x-altus-auth' in request.headers:
            raise Exception("x-altus-auth found in headers!")
        request.headers['x-altus-auth'] = self._get_signature_header(signature)

    def _get_signature_header(self, signature):
        auth_params = OrderedDict()
        auth_params['access_key_id'] = self.credentials.access_key_id
        auth_params['auth_method'] = self.auth_method
        encoded_auth_params = json.dumps(auth_params).encode('utf-8')
        return "%s.%s" % (
            urlsafe_b64encode(encoded_auth_params).strip().decode('utf-8'),
            signature)


class Ed25519v1Auth(V1Signer):
    """
    Ed25519 signing with a SHA-512 hash returning a base64 encoded signature.
    """
    AUTH_METHOD_NAME = 'ed25519v1'
    ED25519_SEED_LENGTH = 32
    ED25519_BASE64_SEED_LENGTH = 44

    def __init__(self, credentials):
        super(Ed25519v1Auth, self).__init__(credentials, self.AUTH_METHOD_NAME)

    @classmethod
    def detect_private_key(cls, key):
        return len(key) == cls.ED25519_BASE64_SEED_LENGTH

    def _raw_sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the raw signature.
        :param string_to_sign: String to sign
        :return: Raw signature as string
        """
        try:
            # We expect the private key to be a base64 formatted string.
            seed = b64decode(self.credentials.private_key)
            if len(seed) != self.ED25519_SEED_LENGTH:
                raise Exception('Not an Ed25519 private key: %s' %
                                self.credentials.private_key)

            pk = eddsa.publickey(seed)
            signature = eddsa.signature(string_to_sign.encode('utf-8'), seed, pk)
            return signature
        except Exception:
            message = self.ERROR_MESSAGE % self.credentials.method
            LOG.debug(message, exc_info=True)
            raise Exception(message)


class RSAv1Auth(V1Signer):
    """
    RSA signing with a SHA-256 hash returning a base64 encoded signature.
    """
    AUTH_METHOD_NAME = 'rsav1'

    def __init__(self, credentials):
        super(RSAv1Auth, self).__init__(credentials, self.AUTH_METHOD_NAME)

    def _raw_sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the raw signature.
        :param string_to_sign: String to sign
        :return: Raw signature as string
        """
        try:
            # We expect the private key to be the an PKCS8 pem formatted string.
            pem_bytes = self.credentials.private_key.encode('utf-8')
            if pem.detect(pem_bytes):
                _, _, der_bytes = pem.unarmor(pem_bytes)
                # In PKCS8 the key is wrapped in a container that describes it
                info = keys.PrivateKeyInfo.load(der_bytes, strict=True)
                # Directly unwrap the private key. The asn1crypto library stopped
                # offering an API call for this in their 1.0.0 release but their
                # official answer of using a separate native-code-dependent
                # library to do one line of work is unreasonable. Of course, this
                # line might break in the future...
                unwrapped = info['private_key'].parsed
                # The unwrapped key is equivalent to pkcs1 contents
                key = rsa.PrivateKey.load_pkcs1(unwrapped.dump(), 'DER')
            else:
                raise Exception('Not a PEM file')
        except Exception:
            message = self.ERROR_MESSAGE % self.credentials.method
            LOG.debug(message, exc_info=True)
            raise Exception(message)
        # We sign the hash.
        signature = rsa.sign(string_to_sign.encode('utf-8'), key, 'SHA-256')
        return signature


class ECDSAv1Auth(V1Signer):
    """
    ECDSA signing with a SHA-512 hash returning a base64 encoded signature.
    """
    AUTH_METHOD_NAME = 'ecdsav1'

    def __init__(self, credentials):
        super(ECDSAv1Auth, self).__init__(credentials, self.AUTH_METHOD_NAME)

    @staticmethod
    def _parse_private_key(private_key, error_message="Failed to import private key."):
        """
        Parse the ECDSA private key.
        :param private_key: The private key.
        :return: The SigningKey.
        :raise: Exception if the private key can not be parsed.
        """
        try:
            # We expect the private key to be a PKCS8 pem formatted string.
            pem_bytes = private_key.encode('utf-8')
            if pem.detect(pem_bytes):
                _, _, der_bytes = pem.unarmor(pem_bytes)
                # In PKCS8 the key is wrapped in a container that describes it
                info = keys.PrivateKeyInfo.load(der_bytes, strict=True)
                # Directly unwrap the private key. The asn1crypto library stopped
                # offering an API call for this in their 1.0.0 release but their
                # official answer of using a separate native-code-dependent
                # library to do one line of work is unreasonable. Of course, this
                # line might break in the future...
                unwrapped = info['private_key'].parsed
                # The unwrapped key is equivalent to pkcs1 contents
                return SigningKey.from_der(unwrapped.dump())
            else:
                raise Exception('Not a PEM file')
        except Exception:
            LOG.debug(error_message, exc_info=True)
            raise Exception(error_message)

    @staticmethod
    def detect_private_key(private_key):
        """
        Try to see if the private key is an ECDSA key by parsing it.
        :param private_key: The key to check.
        :return: True if the key is an ECDSA key, False otherwise.
        """
        try:
            ECDSAv1Auth._parse_private_key(private_key)
            return True
        except Exception:
            return False

    def _raw_sign_string(self, string_to_sign):
        """
        Sign the supplied string using the credentials and return the raw signature.
        :param string_to_sign: String to sign
        :return: Raw signature as string
        """
        sk = ECDSAv1Auth._parse_private_key(
            self.credentials.private_key,
            self.ERROR_MESSAGE % self.credentials.method)
        # Sign the hash.
        signature = sk.sign_deterministic(
            string_to_sign.encode('utf-8'),
            hashfunc=hashlib.sha512,
            sigencode=sigencode_der)
        return signature


class AccessTokenAuth:
    AUTH_METHOD_NAME = 'access_token'

    def __init__(self, credentials):
        self._credentials = credentials

    def add_auth(self, request):
        if self._credentials is None:
            raise NoCredentialsError(
                err_msg='Credentials not available for request signing')
        if 'Authorization' in request.headers:
            raise Exception("Authorization found in headers!")
        request.headers['Authorization'] = self._credentials.access_token

    @classmethod
    def is_access_token(cls, access_token):
        return bool(access_token)


AUTH_TYPE_MAPS = {
    ECDSAv1Auth.AUTH_METHOD_NAME: ECDSAv1Auth,
    Ed25519v1Auth.AUTH_METHOD_NAME: Ed25519v1Auth,
    RSAv1Auth.AUTH_METHOD_NAME: RSAv1Auth,
    AccessTokenAuth.AUTH_METHOD_NAME: AccessTokenAuth
}



# Copyright 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import logging
import os
import urllib.parse as urlparse

from cdpcli.clidriver import CLIOperationCaller
from cdpcli.exceptions import CdpCLIError, DfExtensionError

LOG = logging.getLogger('cdpcli.extensions.df')


def get_expanded_file_path(file_path):
    return os.path.expandvars(os.path.expanduser(file_path))


def upload_workload_asset(client, parameters):
    method = 'post'
    url = '/dfx/api/rpc-v1/deployments/upload-asset-content'
    file_path = parameters.get('filePath', None)
    headers = {
        'Content-Type': 'application/octet-stream',
        'Deployment-Request-Crn': parameters.get('deploymentRequestCrn', None),
        'Deployment-Name': parameters.get('deploymentName', None),
        'Asset-Update-Request-Crn': parameters.get('assetUpdateRequestCrn', None),
        'Parameter-Group': parameters.get('parameterGroup', None),
        'Parameter-Name': parameters.get('parameterName', None),
        'File-Path': file_path,
    }
    return upload_file(client, 'uploadAsset', method, url, headers, file_path)


def upload_file(client, operation_name, method, url, headers, file_path):
    headers = {k: v for k, v in headers.items() if v is not None}
    expanded_file_path = get_expanded_file_path(file_path)
    if os.path.exists(expanded_file_path):
        with open(expanded_file_path, 'rb') as f:
            http, parsed_response = client.make_request(
                operation_name, method, url, headers, f)
    else:
        raise DfExtensionError(
            err_msg='Path [{}] not found'.format(file_path),
            service_name=client.meta.service_model.service_name,
            operation_name=operation_name)
    return parsed_response


class DfExtension(CLIOperationCaller):
    def invoke(self,
               client_creator,
               operation_model,
               parameters,
               parsed_args,
               parsed_globals):
        service_name = operation_model.service_model.service_name
        operation_name = operation_model.name
        if service_name == 'df' and operation_name == 'importFlowDefinition':
            self._df_upload_flow(client_creator,
                                 operation_model,
                                 parameters,
                                 parsed_globals)
        elif service_name == 'df' and operation_name == 'importFlowDefinitionVersion':
            self._df_upload_flow_version(client_creator,
                                         operation_model,
                                         parameters,
                                         parsed_globals)
        elif service_name == 'dfworkload' and operation_name == 'uploadAsset':
            self._df_workload_upload_asset(client_creator,
                                           operation_model,
                                           parameters,
                                           parsed_globals)
        elif service_name == 'dfworkload' and operation_name == 'updateDeployment':
            self._df_workload_update_deployment(client_creator,
                                                operation_model,
                                                parameters,
                                                parsed_globals)
            return True
        else:
            raise DfExtensionError(
                err_msg='The operation is not supported.',
                service_name=service_name,
                operation_name=operation_name)
        # The command processing is finished, do not run the original CLI caller.
        return False

    def _encode_value(self, value):
        if value:
            return urlparse.quote(value)
        return None

    def _df_upload_flow(self, client_creator, operation_model,
                        parameters, parsed_globals):
        client = client_creator('df')
        operation_name = operation_model.name
        url = operation_model.http['requestUri']
        method = 'post'

        # Encode the name, description, and comments fields.
        # The df api expects them to be URI encoded.
        flowName = self._encode_value(parameters.get('name', None))
        flowDescription = self._encode_value(parameters.get('description', None))
        flowComment = self._encode_value(parameters.get('comments', None))

        headers = {
            'Content-Type': 'application/json',
            'Flow-Definition-Name': flowName,
            'Flow-Definition-Description': flowDescription,
            'Flow-Definition-Comments': flowComment
        }

        file_path = parameters.get('file', None)
        response = upload_file(client, operation_name,
                               method, url, headers, file_path)
        self._display_response(operation_name, response, parsed_globals)

    def _df_upload_flow_version(self, client_creator, operation_model,
                                parameters, parsed_globals):
        client = client_creator('df')
        operation_name = operation_model.name
        url = operation_model.http['requestUri']
        method = 'post'

        # Encode the comments fields, the df api expects them to be URI encoded.
        flowComment = self._encode_value(parameters.get('comments', None))
        headers = {
            'Content-Type': 'application/json',
            'Flow-Definition-Comments': flowComment
        }
        file_path = parameters.get('file', None)
        response = upload_file(client, operation_name,
                               method, url, headers, file_path)
        self._display_response(operation_name, response, parsed_globals)

    def _df_workload_upload_asset(self, client_creator, operation_model,
                                  parameters, parsed_globals):
        client = client_creator('dfworkload')
        operation_name = operation_model.name
        response = upload_workload_asset(client, parameters)
        self._display_response(operation_name, response, parsed_globals)

    def _df_workload_update_deployment(self, client_creator, operation_model,
                                       parameters, parsed_globals):
        kpis = parameters.get('kpis', None)
        if kpis:
            self._process_kpis(kpis)

        if self._has_asset_references(parameters):
            df_workload_client = client_creator('dfworkload')

            deployment_crn = parameters.get('deploymentCrn', None)
            environment_crn = parameters.get('environmentCrn', None)
            asset_update_request_crn = self._get_asset_update_request_crn(
                    df_workload_client, environment_crn, deployment_crn)

            LOG.debug('Asset Update Request CRN [%s]' % asset_update_request_crn)
            parameters['assetUpdateRequestCrn'] = asset_update_request_crn

            try:
                self._upload_asset_references(
                        df_workload_client, asset_update_request_crn, parameters)
            except CdpCLIError:
                df_workload_client.abort_asset_update_request(
                    deploymentCrn=deployment_crn,
                    environmentCrn=environment_crn,
                    assetUpdateRequestCrn=asset_update_request_crn
                )
                raise

    def _process_kpis(self, kpis):
        """
        Process Key Performance Indicators and set required unit properties
        """
        for kpi in kpis:
            alert = kpi.get('alert', None)
            if alert:
                frequency_tolerance = alert.get('frequencyTolerance', None)
                if frequency_tolerance:
                    unit = frequency_tolerance['unit']
                    id = unit['id']
                    unit['label'] = id.capitalize()
                    unit['abbreviation'] = id[:1].lower()

    def _has_asset_references(self,
                              parameters):
        """
        Evaluate Parameter Groups and Parameters for Asset References
        """
        parameter_groups = parameters.get('parameterGroups', None)
        if parameter_groups:
            for parameter_group in parameter_groups:
                parameters = parameter_group['parameters']
                for parameter in parameters:
                    asset_references = parameter.get('assetReferences', None)
                    if asset_references:
                        for asset_ref in asset_references:
                            asset_path = asset_ref.get('path', None)
                            if asset_path:
                                return True

        return False

    def _get_asset_update_request_crn(self,
                                      df_workload_client,
                                      environment_crn,
                                      deployment_crn):
        response = df_workload_client.create_asset_update_request(
                deploymentCrn=deployment_crn,
                environmentCrn=environment_crn
        )
        return response.get('assetUpdateRequestCrn', None)

    def _upload_asset_references(self,
                                 df_workload_client,
                                 asset_update_request_crn,
                                 parameters):
        """
        Process Parameter Groups and upload Asset References
        """
        parameter_groups = parameters.get('parameterGroups', None)
        if parameter_groups:
            for parameter_group in parameter_groups:
                parameter_group_name = parameter_group['name']
                parameters = parameter_group['parameters']
                for parameter in parameters:
                    asset_references = parameter.get('assetReferences', None)
                    if asset_references:
                        for asset_reference in asset_references:
                            asset_path = asset_reference.get('path', None)
                            if asset_path:
                                asset_params = {
                                    'assetUpdateRequestCrn': asset_update_request_crn,
                                    'parameterGroup': parameter_group_name,
                                    'parameterName': parameter.get('name', None),
                                    'filePath': asset_path
                                }
                                upload_workload_asset(df_workload_client, asset_params)

                                file_path = get_expanded_file_path(asset_path)
                                path, name = os.path.split(file_path)
                                asset_reference['name'] = name
                                asset_reference['path'] = path
